/***********************************************************************
 @C-file{
    author              = "Nelson H. F. Beebe",
    version             = "4.00",
    date                = "07 July 1998",
    time                = "11:40:41 MDT",
    filename            = "color.c",
    address             = "Center for Scientific Computing
                           University of Utah
                           Department of Mathematics, 322 INSCC
                           155 S 1400 E RM 233
                           Salt Lake City, UT 84112-0090
                           USA",
    telephone           = "+1 801 581 5254",
    FAX                 = "+1 801 585 1640, +1 801 581 4148",
    URL                 = "http://www.math.utah.edu/~beebe",
    checksum            = "04364 836 3079 21135",
    email               = "beebe@math.utah.edu, beebe@acm.org,
                           beebe@ieee.org (Internet)",
    codetable           = "ISO/ASCII",
    keywords            = "color, dired",
    supported           = "yes",
    docstring           = "This file implements the public function

			    const char *colorize(const char *filename)

			   for dired.  colorize() returns a pointer to a
			   static internal copy of filename with
			   surrounding terminal escape sequences to
			   display it in color.

			   Color is supported only for terminal types
			   that are known to be color-capable, defined
			   in the color_terms[] list in is_colorable()
			   below.  For other terminal types, no
			   surrounding escape sequences are inserted.

			   Color choices are provided by built-in
			   defaults in the LS_COLORS value set below (or
			   redefined at compile time). They may be
			   overridden at run-time by a user-defined
			   LS_COLORS (or LS_COLOURS) environment
			   variable, as documented in the accompanying
			   dired.man manual page file.

			   The implementation here is based strictly on
			   the documentation of the GNU fileutils-3.16
			   ls and dircolors programs; the code here is
			   entirely independent of the GNU code.

                           The checksum field above contains a CRC-16
                           checksum as the first value, followed by the
                           equivalent of the standard UNIX wc (word
                           count) utility output of lines, words, and
                           characters.  This is produced by Robert
                           Solovay's checksum utility.",
 }
***********************************************************************/

#include <config.h>

#if defined(HAVE_CTYPE_H)
#include <ctype.h>
#endif

#if defined(HAVE_STDIO_H)
#include <stdio.h>
#endif

#if defined(HAVE_STDLIB_H)
#include <stdlib.h>
#endif

#if defined(HAVE_SYS_TYPES_H)
#include <sys/types.h>
#endif

#if defined(HAVE_UNISTD_H)
#include <unistd.h>		/* for read() et al prototypes */
#endif

#include <stdc.h>		/* private header file */
#include <lbuf.h>		/* private header file */
#include <xstring.h>		/* private header file */

#if !defined(LS_COLORS)
/* data borrowed from GNU fileutils-3.16/src/dircolors.h code */
#define LS_COLORS	"bd=40;33;01:cd=40;33;01:di=01;34:ex=01;32:fi=0:lc=\e[:ln=01;36:mi=:no=0:or=40;31;01:pi=40;33:rc=m:so=01;35"
#undef LS_COLORS
/* data borrowed from GNU fileutils-3.16/src/dircolors output */
#define LS_COLORS	"no=0:fi=0:di=01;34:ln=01;36:pi=40;33:so=01;35:bd=40;33;01:cd=40;33;01:or=40;31;01:ex=01;32:*.tar=01;31:*.tgz=01;31:*.arj=01;31:*.taz=01;31:*.lzh=01;31:*.zip=01;31:*.z=01;31:*.Z=01;31:*.gz=01;31:*.deb=01;31:*.jpg=01;35:*.gif=01;35:*.bmp=01;35:*.ppm=01;35:*.tga=01;35:*.xbm=01;35:*.xpm=01;35:*.tif=01;35:*.mpg=01;37:*.avi=01;37:*.gl=01;37:*.dl=01;37:"
#endif

typedef struct colorcap
{
    const char *key;
    const char *value;
}
ColorCap;

ColorCap colorcap[] =
{
    /*******************************************************************
      Notes:
	(1) a NULL value means that the key was never defined;
	(2) an empty key string means that the slot that can be filled
	    in with new file extension patterns;
	(3) the first empty key string means the end of the table;
	(4) a NUL key string means the end of the table, and also that
	    the table is full;
	(5) the table is reorganized dynamically with a move-to-front
            algorithm by get_cap() to speed its linear search;
	(6) the most-frequently-used entries are stored first, and
	    never moved by get_cap().
    *******************************************************************/

    {"lc", "\033["},			/* Left code */
    {"no", (const char *) NULL},	/* Normal (non-filename) text */
    {"rc", "m"},			/* Right code */
    {"ec", (const char *) NULL},	/* End code (replaces lc+no+rc) */

#define FROZEN_ENTRIES 4		/* index of first movable entry */

    {"bd", (const char *) NULL},	/* Block device */
    {"cd", (const char *) NULL},	/* Character device */
    {"di", (const char *) NULL},	/* Directory */
    {"ex", (const char *) NULL},	/* Executable file */
    {"fi", (const char *) NULL},	/* Regular file */
    {"ln", (const char *) NULL},	/* Symbolic link */
    {"mi", (const char *) NULL},	/* Missing file (defaults to fi) */
    {"or", (const char *) NULL},	/* Orphaned symbolic link (defaults to ln) */
    {"pi", (const char *) NULL},	/* Named pipe (FIFO) */
    {"so", (const char *) NULL},	/* Socket */

    /* Remaining entries may be filled in at runtime with file
       extensions and values */

    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},
    {"", (const char *) NULL},

    /* NULL key marks end of table */
    {(const char *) NULL, (const char *) NULL},
};

/* public functions */
const char		*colorize ARGS ((const char *filename,
					 const struct lbuf * lb));

/* private functions */
static char		*append ARGS ((char *s, size_t maxs, const char *suffix));
static const char	*dirchr ARGS ((const char *s, int c));
static char		*dirtok ARGS ((char *s, const char *sepchars));
static const char	*get_altcap ARGS ((const char *key));
static const char	*get_cap ARGS ((const char *key));
static const char	*get_type ARGS ((const char *filename));
static int		hexval ARGS ((int c_));
static void		initialize ARGS ((void));
static int		is_colorable ARGS ((void));
static void		parse_colors ARGS ((const char *ls_colors));
static void		set_cap ARGS ((const char *key, char *value));
static char		*unescape ARGS ((char *s));


#if STDC
static char *
append (char *s, size_t maxs, const char *suffix)
#else /* NOT STDC */
static char *
append (s, maxs, suffix)
char *s;
size_t maxs;
const char *suffix;
#endif /* STDC */
{				/* return (concat s suffix) if s is big
				   enough, else s */
    size_t len_suffix;
    size_t len_s;

    if (suffix != (const char *) NULL)	/* NULL suffix is same as "" */
    {
	len_s = strlen (s);
	len_suffix = strlen (suffix);
	if ((len_s + len_suffix) < maxs)
	    (void) strcpy (&s[len_s], suffix);
    }
    return (s);
}


#if STDC
const char *
colorize (const char *filename, const struct lbuf *lb)
#else /* NOT STDC */
const char *
colorize (filename, lb)
const char *filename;
const struct lbuf *lb;
#endif /* STDC */
{	/* return a pointer to a static internal buffer with a colorized
	   copy of filename, or to filename, if no colorization is done */
    static char buffer[2 * MAX_DNAME];	/* allow plenty of extra space for
					   escape sequences */
    const char *key;

    initialize ();

    if (is_colorable())
    {
	if ((key = get_type (filename)) == (const char *) NULL)
	{
	    switch (lb->ltype)
	    {
	    case 'D':			/* treat door like socket */
		key = "so";
		break;

	    case 'b':			/* block device */
		key = "bd";
		break;

	    case 'c':			/* character device */
		key = "cd";
		break;

	    case 'd':			/* directory */
		key = "di";
		break;

	    case 'l':			/* symbolic link */
		key = "ln";
		break;

	    case 'p':			/* pipe */
		key = "pi";
		break;

	    case 's':			/* socket */
		key = "so";
		break;
					    /* normal file */
	    case '-':
	    default:			/* everything else */
		key = "fi";
		break;
	    }
	}

	/* From the GNU fileutils-3.16 ls manual page:

	   Each file will be written as
	   <lc><color code><rc><filename><ec>
	   If the <ec> code is undefined, the sequence <lc><no><rc>
	   will be used instead.
	 */

	buffer[0] = '\0';
	append (buffer, sizeof (buffer), get_altcap ("lc"));
	append (buffer, sizeof (buffer), get_altcap (key));
	append (buffer, sizeof (buffer), get_altcap ("rc"));
	append (buffer, sizeof (buffer), filename);
	append (buffer, sizeof (buffer), get_altcap ("ec"));

	return ((const char *) &buffer[0]);
    }
    else
	return (filename);
}


#if STDC
static const char *
dirchr (const char *s, int c)
#else /* NOT STDC */
static const char *
dirchr (s, c)
const char *s;
int c;
#endif /* STDC */
{
    /* dirchr() is like strchr(), except that c preceded by backslash is
       not recognized */

    const char *t;

    for (t = s; *t; ++t)
    {
	if (*t == c)
	{
	    if (t == s)
		return (t);
	    else if ((t > s) && (t[-1] != '\\'))
		return (t);
	}
    }
    return (((c == 0) && (*t == '\0')) ? t : (const char *) NULL);
}


#if STDC
static char *
dirtok (
	   char *s,		/* non-NULL on first call, NULL on subsequent ones */
	   const char *sepchars	/* separator characters */
)
#else /* NOT STDC */
static char *
dirtok (s, sepchars)
char *s;			/* non-NULL on first call, NULL on subsequent ones */
const char *sepchars;		/* separator characters */
#endif /* STDC */
{
    /* dirtok() is a simple implementation of strtok(), with
       modifications to support backslash quoting of separator
       characters to prevent them from acting as separators.  */

    static char *s_init = (char *) NULL;
    char *start;
    char *end;

    if (s != (char *) NULL)	/* initial call for this s[] */
	s_init = s;
    else if ((s_init == (char *) NULL) || (*s_init == '\0'))
	return ((char *) NULL);

    /* resume scan from saved token start */

    for (start = s_init;
	 *start && dirchr (sepchars, *start) != (const char *) NULL;
	 start++)
	/* skip over leading separators */ ;
    if (*start == '\0')
	return ((char *) NULL);
    for (end = start;
	 *end && dirchr (sepchars, *end) == (const char *) NULL; end++)
	/* skip over token */ ;
    s_init = *end ? end + 1 : end;	/* next token starts past this */
    /* but not beyond end of string */
    *end = '\0';		/* terminate current token */
    return (start);		/* and return pointer to it */
}


#if STDC
static const char *
get_altcap (const char *key)
#else /* NOT STDC */
static const char *
get_altcap (key)
const char *key;
#endif /* STDC */
{
    const char *value;

    /*******************************************************************
       For uncear reasons, the GNU ls color support defines alternates
       for certain capabilities when they are undefined:

		ec -> lc no rc
		mi -> fi
		or -> ln
    *******************************************************************/

    if (strcmp (key, "ec") == 0)
    {
	value = get_cap (key);
	if (value == (const char *) NULL)
	{
	    static char buffer[MAX_DNAME];

	    buffer[0] = '\0';
	    (void) append (buffer, sizeof (buffer), get_cap ("lc"));
	    (void) append (buffer, sizeof (buffer), get_cap ("no"));
	    value = append (buffer, sizeof (buffer), get_cap ("rc"));
	}
    }
    else if (strcmp (key, "mi") == 0)
    {
	value = get_cap (key);
	if (value == (const char *) NULL)
	    value = get_cap ("fi");
    }
    else if (strcmp (key, "or") == 0)
    {
	value = get_cap (key);
	if (value == (const char *) NULL)
	    value = get_cap ("ln");
    }
    else
	value = get_cap (key);

    return (value);
}


#if STDC
static const char *
get_cap (const char *key)
#else /* NOT STDC */
static const char *
get_cap (key)
const char *key;
#endif /* STDC */
{
    size_t k;

    for (k = 0; colorcap[k].key != (const char *) NULL; ++k)
    {
	if (colorcap[k].key[0] == '\0')	/* reached end of defined keys */
	    return ((const char *) NULL);
	else if (strcmp (colorcap[k].key, key) == 0)
	{
	    if (k > FROZEN_ENTRIES)
	    {
		ColorCap swap;

	        /* Optimization: swap the entry found and the first
		   unfrozen entry */
		swap = colorcap[k];
		colorcap[k] = colorcap[FROZEN_ENTRIES];
		colorcap[FROZEN_ENTRIES] = swap;
		k = FROZEN_ENTRIES;
	    }
	    return (colorcap[k].value);
	}
    }
    return ((const char *) NULL);
}


#if STDC
static const char *
get_type (const char *filename)
#else /* NOT STDC */
static const char *
get_type (filename)
const char *filename;
#endif /* STDC */
{
    const char *p;

    p = strrchr (filename, '.');
    if (p != (const char *) NULL)
    {
	static char buffer[MAX_DNAME];

	buffer[0] = '\0';
	append (buffer, sizeof (buffer), "*");
	append (buffer, sizeof (buffer), p);
	p = &buffer[0];
    }
    return (p);
}


#if STDC
static int
hexval (int c)
#else /* NOT STDC */
static int
hexval (c)
int c;
#endif /* STDC */
{
    static const char *hexdigits = "0123456789abcdef";
    const char *p = (const char *) strchr (hexdigits,
					   isupper (c) ? tolower (c) : c);

    return ((p == (const char *) NULL) ? (-1) : (int) (p - hexdigits));
}


static void
initialize (VOID_ARG)
{
    static int first = 1;

    if (first)
    {
	const char *p;

	if (is_colorable())
	{
	    p = getenv ("LS_COLORS");
	    if (p == (const char *) NULL)
		getenv ("LS_COLOURS");
	    parse_colors ((p == (const char *) NULL) ? LS_COLORS : p);
	}
	first = 0;
    }
}


static int
is_colorable (VOID_ARG)
{
    static const char *color_terms[] =
    {				/* this list is borrowed from GNU
				fileutils-3.16/src/dircolors.h */
	"xterm",
	"vt100",
	"linux",
	"console",
	"con80x60",
	"con80x50",
	"con80x43",
	"con80x30",
	"con80x28",
	"con80x25",
	"con132x60",
	"con132x43",
	"con132x30",
	"con132x25",
	(const char *) NULL
    };
    static int first = 1;
    static int cached_value = 0;
    const char *term;
    size_t k;

    if (first)
    {
	cached_value = 0;
	term = getenv ("TERM");
	if (term != (const char *) NULL)	/* then TERM is set */
	{
	    for (k = 0; color_terms[k] != (const char *) NULL; ++k)
	    {
		if (strcmp (color_terms[k], term) == 0)
		{
		    cached_value = 1;
		    break;
		}
	    }
	}
	first = 0;
    }
    return (cached_value);
}


#if STDC
static void
parse_colors (const char *ls_colors)
#else /* NOT STDC */
static void
parse_colors (ls_colors)
const char *ls_colors;
#endif /* STDC */
{
    size_t n;
    char *temp_ls_colors;
    char *p;

    if (ls_colors == (const char *) NULL)
	return;

    /* We allocate, but never free, a copy of ls_colors, because we
       need to retain pointers into its pieces in the colorcap[] table */
    n = strlen (ls_colors);
    temp_ls_colors = (char *) malloc (n + 1);
    if (temp_ls_colors == (const char *) NULL)
	return;

    /* ls_colors looks like
       key1=value1:key2=value2:key3=value3:key4=value4:... */

    for (p = dirtok (strcpy (temp_ls_colors, ls_colors), ":");
	 p != (char *) NULL; p = dirtok ((char *) NULL, ":"))
    {
	char *equals;

	equals = strchr (p, '=');
	if (equals != (char *) NULL)
	{
	    *equals = '\0';
	    set_cap (p, equals + 1);
	}
    }
}


#if STDC
static void
set_cap (const char *key, char *value)
#else /* NOT STDC */
static void
set_cap (key, value)
const char *key;
char *value;
#endif /* STDC */
{
    size_t k;

    for (k = 0; colorcap[k].key != (const char *) NULL; ++k)
    {
	if (colorcap[k].key[0] == '\0')
	    break;		/* reached end of key-value definitions */
	if (strcmp (colorcap[k].key, key) == 0)
	{
	    colorcap[k].value = unescape (value);
	    return;
	}
    }
    if (colorcap[k].key != (const char *) NULL)
    {					/* then have space to store this pair */
	colorcap[k].key = key;
	colorcap[k].value = unescape (value);
    }
}


#if STDC
static char *
unescape (char *s)
#else /* NOT STDC */
static char *
unescape (s)
char *s;
#endif /* STDC */
{
    /* Replace escape sequences in s in-line and return a pointer to it. */
    /* Code borrowed from bibunlex.c, then modified for LS_COLORS extensions */

#define digval(c)	((int)((c) - '0'))
#define isoctal(c)	(isdigit(c) && ((c) != '8') && ((c) != '9'))
#define ishexadecimal(c) (isdigit(c) || \
				(strchr("ABCDEFabcdef",(c)) != (char*)NULL))

    int c;
    int n;
    char *original = s;
    char *t;			/* target string (overwrites s[]) */

    /*******************************************************************
      From the GNU fileutils-3.16 ls manual page:

	Control characters can be written either in C-style \- escaped
	notation, or in stty-like ^-notation.  The C-style notation
	adds \e for Escape, \_ for a normal space characer, and \? for
	Delete.  In addition, the \ escape character can be used to
	override the default interpretation of \, ^, : and =.
    *******************************************************************/

    for (t = s; *s;)
    {
	switch (*s)
	{
	case '\\':
	    ++s;
	    switch (*s)
	    {
	    case '0':
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
		for (n = 0, c = digval (*s), ++s; (n < 3) && isoctal (*s);
		     n++, s++)
		    c = 8 * c + digval (*s);
		break;

	    case 'a':
		++s;
		c = '\007';	/* '\a' */
		break;

	    case 'b':
		++s;
		c = '\b';
		break;

	    case 'f':
		++s;
		c = '\f';
		break;

	    case 'n':
		++s;
		c = '\n';
		break;

	    case 'r':
		++s;
		c = '\r';
		break;

	    case 't':
		++s;
		c = '\t';
		break;

	    case 'v':
		++s;
		c = '\v';
		break;

	    case 'x':
		for (++s, c = hexval (*s); ishexadecimal (*s); ++s)
		    c = 16 * c + hexval (*s);
		break;

	    case 'e':		/* \e: special addition for LS_COLORS support */
		++s;
		c = '\033';
		break;

	    case '_':		/* \_: special addition for LS_COLORS support */
		++s;
		c = ' ';
		break;

	    case '?':		/* \?: special addition for LS_COLORS support */
		++s;
		c = '\177';
		break;

	    default:
		c = *s++;
	    }
	    *t++ = c;
	    break;

	case '^':		/* ^x: special addition for LS_COLORS support */
	    *t++ = 037 & *(++s);	/* controlify following character */
	    break;

	default:
	    *t++ = *s++;
	    break;
	}
    }
    *t = '\0';
    return (original);
}
