/*
 **********************************************************************
 *                                                                    *
 * dired [][dir][files]  Stuart Cracraft (mclure@sri-unix) Sept 1980  *
 *                                                                    *
 *      Directory editor.                                             *
 *              edit/delete files in a directory or a file list       *
 *              compile with the berkeley termlib archive             *
 *              and don't forget to change the 'helpfile' and         *
 *              'dirednam' strings to reflect your own setup.         *
 *                                                                    *
 *      Note: if you make improvements, I'd like to get them too.     *
 *                      Stuart Cracraft mclure@sri-unix,              *
 *                              ucbvax!menlo70!sri-unix!mclure        *
 *      so would I:     Jay Lepreau     lepreau@utah-20,              *
 *                      decvax!{harpo,randvax}!utah-cs!lepreau        *
 *      me too:         Charles Hill     philabs!crh                  *
 **********************************************************************
 */

/*
 *    Enhanced by J.Lepreau, Univ of Utah, 8-10/81:
 *            --bunch of stuff, including more cmds, 2 window mode,
 *              bufering output, pathname substition in !escapes,
 *              initial sort options, ^G escape from sorts, ^U escape
 *              from !escapes, empty dir check.
 *


 Modified 12/81 by Lepreau:
 1. Fix up aborting of command escapes, with mildy tricky re-display.
 2. Make default window size be half-screen (2 window mode).
 Added option -wf to get full screen.
 3. Add check for too many files to avoid core-dump.
 4. Fix bug in cmd escapes: must reset caught signals before invocation.

 Modified 6/82 by Charles Hill (crh)
 Philips Labs
 345 Scarborough Rd.
 Briarcliff Manor, N.Y. 10510
 philabs!crh

 1. Eliminated need for terminal home capability. (Many termcap entries
 leave this out when there is cm). Changed body of home ()
 routine to a call to curxy(0,0).
 2. Made default editor "vi".
 3. On exit, screen is not blanked; cursor moves to last line. Line is
 cleared
 4. In "type", only outputs 79 chars/line so terms with automatic margins
 will work ok.
 5. Changed type () to handle tabs better.
 6. Changed ceol to work for terms w/o clear to eoln. Single call to ceod
 eliminated. If no clear to eod, this is simulated. But it's better to
 use full screen on such terminals.
 7. Changed type () to check if a file is a text file first. On some
 terminals, outputting arbitrary binary can put them in a weird mode,
 and screw up the display.

 Additional enhancements 8/82 crh
 1. Added pattern matching in the form of /, n, and N commands.
 2. Added toggle split-screen command \.
 3. Added p command to print path name.
 4. Change to "c" shell for ! command.
 5. Fixed bug for single file arg when not a dir.
 6. Added 'A', 'Q' commands and "linear window indicator".
 7. Added counts to <up>, <down>, f, and b commands.
 8. Added "G" as in "vi".
 9. Added "E" command to go up a dir, and code to print node name.
 10.Added "R" command to re-read dir.

 ****** 4.2 version 8/83 Charles Hill (crh) *****

 The 4.2 version is shorter because 4.2 has system routines opendir,
 readir, etc. and -r option of rm

 fixes 1/84 crh
 1) show embedded control chars
 2) eliminate 1 line windows except where necessary
 3) cursor bug on deleting last file

 *********************************

 To compile: cc -o dired dired.c cshsystem.o -ltermcap /a/gam/lib/lib.a
 ^ for getshell()
 */


/*
   things to consider
   -    perhaps should be able to edit protection field and
   be able to change protection instantly at the
   touch of a key (or queue the change for later exit)

   --   Would be nice if there were cmds (e.g. 'c') which would update
   1) the file-info & re-display it, & 2) update whole
   display's info.
 */

#if 0
#define	False		0
#define	True		1
typedef int		Boolean;
#else
typedef enum { False = 0, True = 1 } Boolean;
#endif

#include <config.h>

#if defined(HAVE_CTYPE_H)
#include <ctype.h>
#endif

#if defined(HAVE_LIMITS_H)
#include <limits.h>
#if !defined(INT_MAX)
#define INT_MAX	32767
#endif
#endif

#if defined(HAVE_STDIO_H)
#include <stdio.h>
#endif

#if defined(HAVE_STDLIB_H)
#include <stdlib.h>
#endif

#if defined(HAVE_SYS_TYPES_H)
#include <sys/types.h>
#endif

#if defined(HAVE_TIME_H)
#include <time.h>
#endif

#if defined(HAVE_UNISTD_H)
#include <unistd.h>		/* for read() et al prototypes */
#endif

#if defined(HAVE_SYS_IOCTL_H)
#include <sys/ioctl.h>		/* for ioctl() on some systems */
#endif

#if defined(HAVE_SYS_43IOCTL_H)
#include <sys/43ioctl.h>	/* for TIOxxxxx */
#endif

#if defined(HAVE_GRP_H)
#include <grp.h>
static int have_gid = False;	/* True when -ggid-or-groupname has been given */
static gid_t Gid = 0;
#endif

#if defined(HAVE_PWD_H)
#include <pwd.h>
static int have_uid = False;	/* True when -uuid-or-username has been given */
static uid_t Uid = 0;
#endif

#if defined(HAVE_FCNTL_H)
#include <fcntl.h>
#endif

/* Even though we should prefer POSIX <termios.h> over System V
   <termio.h> or BSD <sgtty.h>, on Sun Solaris 2.x, <curses.h> expects
   <termio.h> to be included too, sigh... */

#if defined(HAVE_TERMIOS_H)
#include <termios.h>
#define USE_TERMIOS
#undef HAVE_SGTTY_H
#endif

#if defined(HAVE_TERMIO_H)
#define USE_TERMIO
#undef HAVE_SGTTY_H
#include <termio.h>		/* needed on Sun Solaris 2.3 */
					/* must come AFTER <termios.h> on HP-UX */
#endif

#if defined(HAVE_SGTTY_H)
#define USE_SGTTY
#include <sgtty.h>		/* must come AFTER <curses.h> */
#endif

/*  We use only the first found of the termios, termio, and sgtty interfaces. */

#if defined(USE_TERMIOS)
#undef USE_TERMIO
#undef USE_SGTTY
static struct termios tty_save;	/* XPG3, POSIX.1, FIPS 151-1 interface */
#endif

#if defined(USE_TERMIO)
#undef USE_SGTTY
static struct termio tty_save;	/* System V interface */
#endif

#if defined(USE_SGTTY)
static struct sgttyb ioctlb;
#endif

#if defined(USE_TERMIOS)
#define KILL_CHAR	tty_save.c_cc[VKILL]
#define ERASE_CHAR	tty_save.c_cc[VERASE]
#endif

#if defined(USE_TERMIO)
#define KILL_CHAR	tty_save.c_cc[VKILL]
#define ERASE_CHAR	tty_save.c_cc[VERASE]
#endif

#if defined(USE_SGTTY)
#define KILL_CHAR	ioctlb.sg_kill
#define ERASE_CHAR	ioctlb.sg_erase
#endif

#if defined(HAVE_SYS_SYSMACROS_H)
#include <sys/sysmacros.h>	/* for major(), minor() prototypes */
#endif

#if defined(HAVE_SYS_MKDEV_H)
#include <sys/mkdev.h>		/* for major(), minor() prototypes */
#endif

#if defined(__cplusplus) || defined(c_plusplus)
extern "C"
{
#endif

#if defined(_AIX) || defined(__linux__) || defined(__hpux) || 1
    /* major and minor are macros on these systems, but configure cannot
       detect that */
#else
#if !defined(HAVE_MAJOR_DECLARATION)
    major_t major ARGS ((const dev_t));
#endif

#if !defined(HAVE_MINOR_DECLARATION)
    minor_t minor ARGS ((const dev_t));
#endif
#endif

#if defined(__cplusplus) || defined(c_plusplus)
}
#endif

#if !defined(HAVE_GETWD)
#undef getwd
#define getwd(p)	getcwd(p,sizeof(p))
#endif

#if defined(HAVE_SYS_PARAM_H)
#include <sys/param.h>
#endif

#if defined(HAVE_SYS_STAT_H)
#include <sys/stat.h>
#endif

#if defined(HAVE_DIRENT_H)
#include <dirent.h>
#define D_NAMLEN(dp)		(int)strlen((dp)->d_name)
#else
#if defined(HAVE_SYS_DIR_H)
#include <sys/dir.h>
#endif

#define dirent direct		/* we use dirent in rest of code */

#if defined(HAVE_SYSTEM7_DIR_H)
/* I don't think any systems with <sys/dir.h> but no d_namlen entry
   exist any more, but if you have one, this should do the trick. */
#define D_NAMLEN(dp)	14	/* really old UNIX System 7 style: */
					/* d_name[] that might lack a NUL */
#else
#define D_NAMLEN(dp)	(int)((dp)->d_namlen)	/* newer <sys/dir.h> style */
					/* contains explicit length field */
#endif
#endif				/* defined(HAVE_DIRENT_H) */


#define DIRPBLOCK	(BUFSIZ / sizeof(struct dirent))
				/* number of directory entries per block */

#if !defined(HAVE_UNLINK) && defined(HAVE_REMOVE)
#define unlink remove
#endif

#include <stdc.h>		/* private header file */
#include <lbuf.h>		/* private header file */
#include <xsignal.h>		/* private header file */
#include <xstring.h>		/* private header file */
#include <xterm.h>		/* private header file */
#include <xwait.h>		/* private header file */

#if defined(HAVE_REGEX_H)
/* IEEE Std 1003.1-2001 (POSIX) regular-expression library */
#include <regex.h>
#define re_comp(pattern)	((regcomp)(&preg, pattern, REG_NOSUB) ? "parse failure" : (char*)NULL)
#define re_exec(pattern)	((regexec)(&preg, pattern, 0, (regmatch_t*)NULL, 0) == 0)
static regex_t preg;
#else
/* Use old BSD-style re_comp() and re_exec(), from private implementation */
#include <dregexpr.h>		/* private header file */
#endif

#if defined(__cplusplus) || defined(c_plusplus)
extern "C"
{
#endif

#if !defined(HAVE_IOCTL_DECLARATION)
    extern int ioctl ARGS ((int, int,...));
#endif

#if defined(__cplusplus) || defined(c_plusplus)
}
#endif

#define ABORTSTAT	69

#if defined(CTRL)
#undef CTRL
#endif
#define CTRL(c)		(c & 037)

#if defined(USE_LS_DATE)
/***********************************************************************
   File date is old confusing ls-style (12 chars) with different
   formats for files younger than, and older than, six months:
	"Jul  4 12:32"
	"Oct  9  1997"
   They are extracted from the ctime() return value: "Sat Jul  4 13:16:03 MDT 1998"

   Typical normal line:
   -rwxr-xr-x 2 beebe          118212 Aug 28  1997  gcc
   00000000001111111111222222222233333333334444444444555555555566666666667777777777
   0123456789.123456789.123456789.123456789.123456789.123456789.123456789.123456789

   Typical line with inode:
	435401 -rwxr-xr-x 2 beebe          118212 Aug 28  1997  gcc
   00000000001111111111222222222233333333334444444444555555555566666666667777777777
   0123456789.123456789.123456789.123456789.123456789.123456789.123456789.123456789
***********************************************************************/
#define CURSOR          (48 + ((sortyp == INODE) ? 12 : 0))
					/* X-coord of cursor, just before filename */
#else
/***********************************************************************
 File date is more informative ISO style (19 chars):
	"1998.07.04 12:32:59"

   Typical normal line:
   -rwxr-xr-x 2 beebe          118212 1997.08.28 09:03:41  gcc
   00000000001111111111222222222233333333334444444444555555555566666666667777777777
   0123456789.123456789.123456789.123456789.123456789.123456789.123456789.123456789

   Typical line with inode:
	435401 -rwxr-xr-x 2 beebe          118212 1997.08.28 09:03:41  gcc
   00000000001111111111222222222233333333334444444444555555555566666666667777777777
   0123456789.123456789.123456789.123456789.123456789.123456789.123456789.123456789

***********************************************************************/
#define CURSOR          (7 + 48 + ((sortyp == INODE) ? 12 : 0))
					/* X-coord of cursor, just before filename */
#endif

#undef CURSOR
#define CURSOR	cursor_x		/* dired 4.05: dynamic cursor position */

#if !defined(DEFAULT_EDITOR)
#define DEFAULT_EDITOR	"vi"
#endif

#define	DELETED		1

#if !defined(DIREDFILE)
#define DIREDFILE	"/usr/local/bin/dired"
#endif

#define DIVCHAR	'-'		/* makes up the divider */

#define ESC         '\033'

#define FNAME(n)	(arglist ? file[n].ln.namep : file[n].ln.lname)

#undef FPRINTF
#define FPRINTF	(void)fprintf

#if !defined(HELPFILE)
#define HELPFILE	"/usr/local/bin/dired.hlp"
#endif

#define	ISARG		0100000

#undef MAX
#define MAX(a,b)	((a) > (b) ? (a) : (b))

#undef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))

#if !defined(MAXFILEPATHLEN)	/* can override at compile-time */
#define MAXFILEPATHLEN	MAX_DNAME /* longest file path name on any */
				/* current UNIX system; called */
				/* FNMAX, MAXPATH, _MAX_PATH, */
				/* MAXPATHLEN, PATH_MAX, and */
				/* FILENAME_MAX on various systems */
#endif

#if !defined(MAXLOOKAHEAD)
#define MAXLOOKAHEAD	1024	/* how far to scan in checking for binary file */
#endif

#if !defined(MAXMARKS)		/* can override at compile-time */
#define MAXMARKS	1024
#endif

#if !defined(MAXVIEW)		/* can override at compile-time */
#define MAXVIEW		10240
#endif

#if !defined(MAXWIDTH)		/* can override at compile-time */
#define MAXWIDTH 255		/* widest lines we will use */
#endif

#if !defined(MOREPGM)		/* can override at compile time */
#if defined(__convex__)
#define MOREPGM	"/usr/ucb/more"	/* program to page thru a file */
#else
#define MOREPGM	"/usr/local/bin/less"	/* program to page thru a file */
#endif /* MOREPGM */
#endif /* defined(__convex__) */

#define	NOTDELETED	(!(DELETED))

#undef PRINTF
#define PRINTF	(void)printf

#undef SPRINTF
#define SPRINTF	(void)sprintf

#define	STRING_EQUAL(a, b)	(strcmp(a, b) == 0)

#define UNSET_GID_OR_UID	(-1L)

static Boolean batch_mode = False; /* non-interactive mode */
static Boolean camedown = False; /* True: -d argument set (recursive dired) */
static Boolean iflg = False;	/* True: show link count in output */
static Boolean intype = False;
static Boolean lflg = True;
static Boolean monochrome = False; /* True: suppress filename colorization */
static Boolean quitok = True;
static Boolean sflg = False;	/* True: show block count in output */
static Boolean show_link_target = True;
static Boolean sigint;
static Boolean updateWinsize = False;	/* GBG 1996, schedule window size update */
static Boolean wasspecial = False;
static Boolean wasstartfile = False;

static char bufout[BUFSIZ];	/* MUST be global; see ANSI C 1989 section 4.9.5.5 */
static char divider[MAXWIDTH + 1];
static char fullpath[MAXFILEPATHLEN];
static char nambuf[MAXFILEPATHLEN];
static char newdivider[MAXWIDTH + 1]; /* divides the windows */
static char startarg[MAXFILEPATHLEN];
static char startfile[MAXFILEPATHLEN];
static char tmpbuf2[MAXFILEPATHLEN];
static char tmpbuf[MAXFILEPATHLEN];	/* Random temporary buffer */

/* The termcap string names have private prefix "my_" in dired 4.00 and
   later, to avoid conflicts with symbols in libtermlib.a and
   libcurses.a on some systems. */
static char *my_AL;		/* insert line */
static char *my_CD;		/* clear to end of display */
static char *my_CE;
static char *my_CL;
static char *my_CM;		/* clear memory */
static char *my_UP;

#if 0
static char *my_HO;
#endif

#if 0
static char my_PC;
#endif


#ifdef V4P2
static char wd[1024];
#else
static int off;
static char name[BUFSIZ];
static char wd[256];
#endif

static const char *dirednam = DIREDFILE;	/* Where dired is */
static const char *helpfile = HELPFILE;		/* Where helpfile lives */

static int LW;			/* the last pos of the linear window */
static int Tscrlen;		/*Total length of screen,minus 1 for cmd line */
static int Worktop = 0;		/* Top of    " " */
static int arglist = 0;
static int asswin = -1;
static int blurb;		/* 1=>thing in echo area,0 otherwise  */
static int command;		/* Holds last command */
static int countarg;
static int curfile = 0;		/* Current file */
static int curline = 0;		/* Line that we're on */
static int cursor_x = 0;	/* Horizontal screen position of deletion flag */
static int errcode = 0;		/* Error variable used by rm */
static int flags;		/* Gets flags */

static int m1[] =
{
    1, S_IREAD >> 0, 'r', '-'
};

static int m2[] =
{
    1, S_IWRITE >> 0, 'w', '-'
};

static int m3[] =
{
    2, S_ISUID, 's', S_IEXEC >> 0, 'x', '-'
};

static int m4[] =
{
    1, S_IREAD >> 3, 'r', '-'
};

static int m5[] =
{
    1, S_IWRITE >> 3, 'w', '-'
};

static int m6[] =
{
    2, S_ISGID, 's', S_IEXEC >> 3, 'x', '-'
};

static int m7[] =
{
    1, S_IREAD >> 6, 'r', '-'
};

static int m8[] =
{
    1, S_IWRITE >> 6, 'w', '-'
};

static int m9[] =
{
    2, S_ISVTX, 't', S_IEXEC >> 6, 'x', '-'
};

static int *m[] =
{
    m1, m2, m3, m4, m5, m6, m7, m8, m9
};

static int max_gname_width = 0;
static int max_gid_width = 0;
static int max_lsize_width = 0;
static int max_uname_width = 0;
static int max_uid_width = 0;

static int numarg;
static int numdeleted;		/* Number of files marked deleted */
static const char *prog;	/* program name (set in main()) */
static int rflg = 1;		/* Sort direction: +1: normal, -1: reverse */
static int scrlen = 0; 		/* Default length of dired index part of screen: */
				/* 999 ==> 2 windows (half size), 0 ==> 1 window (full size) */
static int scrwid = 79;		/* Width of screen */
static int specialcc;
static int splitflg;		/* Split screen? */
static int topfile;		/* File at top of screen */
static size_t totfiles = (size_t)0; /* Total files */
static size_t maxtotfiles = (size_t)0; /* current size of dynamic file[] array */

#define FILE_SLOTS_TO_ADD	((size_t)((65536L - 64L) / sizeof(struct lbuf)))
				/* chunk size of about one 64KB Intel x86 segment */
				/* for optimal memory allocation */

static int view_stack[MAXVIEW];
static int view_top;

#if 0
static int Worklen = 0;		/* Length of 'working window', the other part */
#endif

static long totblocks = 0;	/* Total blocks */

static struct lbuf *file = (struct lbuf *) NULL; /* dired 4.00: big array of
			structures,  file[], is now allocated dynamically */

#if !defined(HAVE_ISATTY)
#define isatty(fd)	((fd), True) /* so that we default to interactive mode */
#endif

#if defined(HAVE_VOID_HANDLER)
#define handler_t void
#else
#define handler_t int
#endif

extern int		cshsystem ARGS ((const char *s));
extern const char	*colorize ARGS ((const char *filename, const struct lbuf * lb));

int			main ARGS ((int argc, char *argv[]));

static void		blank ARGS ((void));
static char		*bldnam ARGS ((char *str, int narg, int filidx, char *argv[]));
static void		bscreen ARGS ((int count));
static char		*catargs ARGS ((char *str, char *argv[]));
static handler_t	catchint ARGS ((int sig));
static void		ceol ARGS ((int x, int y));
static void		curxy ARGS ((int col, int lin));
static void		dcanon ARGS ((char *cp));
static void		deletefiles ARGS ((int argc, char *argv[]));
static void		dispfile ARGS ((int fnum));
static void		dispwin ARGS ((int filenum));
static void		downline ARGS ((void));
static void		expandpath ARGS ((char pname[]));
static void		findfile ARGS ((int down));
static Boolean		findfname ARGS ((char filename[], int *num));
static void		formstartarg ARGS ((char *name));
static void		fscreen ARGS ((int count));
static void		Getcap ARGS ((void));
static int		Getch ARGS ((FILE * fd));
static void		getcountarg ARGS ((void));
static void		getfiles ARGS ((int argc, char *argv[]));
static char		*getfullpath ARGS ((const char *start));
static int		getline ARGS ((char s[]));
static int		getname ARGS ((Boolean want_group, long gid_or_uid, char buf[], size_t maxbuf));
static void		getgids ARGS ((void));
static void		getuids ARGS ((void));
static void		getWinSize ARGS ((void));
static void		grow_file_array ARGS ((void));
static int		gstat ARGS ((const char *gname));
static void		help_and_exit ARGS ((void));
static void		home ARGS ((void));
static void		insline ARGS ((void));
static void		locstart ARGS ((char strtfile[]));
static char		*makename ARGS ((const char *dir, const char *filen, int len));
static int		match ARGS ((int filenum));
static void		my_exit ARGS ((int status));
static long		nblock ARGS ((long size));
static Boolean		nottext ARGS ((const char fn[]));
static void		off_signals ARGS ((void));
static handler_t	onctlz ARGS ((int sig));
static void		on_signals ARGS ((void));

#if defined(TPUTS_HAS_FUNCTION_WITH_CHAR_ARG)
#define OUTCH_ARGTYPE	char
#else
#define OUTCH_ARGTYPE	int
#endif

static int		outch ARGS ((OUTCH_ARGTYPE c));
static void		overflow ARGS ((void));
static void		pentry ARGS ((int whichone));
static void		pmode ARGS ((int aflag));
static void		printpath ARGS ((char str[], int printbottom));
static void		prtsplitstr ARGS ((const char str[]));
static void		putpad ARGS ((char *str));
static void		putstr ARGS ((char *s, int *pcur_x));

static void		Readdir ARGS ((const char *dir));

static void		refresh_screen ARGS ((void));
static void		rm ARGS ((char arg[]));

#ifndef V4P2
static int		Rmdir ARGS ((char *f));
#endif

static void		Select ARGS ((int *pairp));
static void		setdpy ARGS ((void));
static void		set_sorttyp ARGS ((int direction, int type));
static void		shortcountmess ARGS ((void));
static void		showscreen ARGS ((void));
static char		*skipto ARGS ((const char *string, const char *charset));
static void		telluser ARGS ((const char *msg, const char *arg));
static Boolean		type ARGS ((const char *filestr, Boolean waitflg));
static void		typefiles ARGS ((void));
static void		typein ARGS ((const char *argv));
static void		unsetdpy ARGS ((void));
static int		waitchk ARGS ((Boolean waitflg));
static void		windsize ARGS ((void));

#if defined(SIGWINCH)
static handler_t	win_change ARGS ((int sig));
#endif

#if 0
static void		ceod ARGS ((void));
static void		upline ARGS ((void));
#endif

#ifndef V4P2
static int		dotname ARGS ((char *s));
static void		prname ARGS ((void));
#endif

/* Sort Orders:  */
#define CREATE		0
#define GIDNAME		1
#define GIDNUMBER	2
#define INODE		3
#define LINKS		4
#define NAME		5
#define NUMBER		6
#define READ		7
#define SIZE		8
#define TYPE		9
#define UIDNAME		10
#define UIDNUMBER	11
#define WRITE		12

static int sortyp;		/* Key to sort on: CREATE..WRITE */

static int		cmp_create	ARGS ((const void *pp1, const void *pp2));
static int		cmp_gid		ARGS ((const void *pp1, const void *pp2));
static int		cmp_gid_number	ARGS ((const void *pp1, const void *pp2));
static int		cmp_inode	ARGS ((const void *pp1, const void *pp2));
static int		cmp_links	ARGS ((const void *pp1, const void *pp2));
static int		cmp_name	ARGS ((const void *pp1, const void *pp2));
static int		cmp_number	ARGS ((const void *pp1, const void *pp2));
static int		cmp_read	ARGS ((const void *pp1, const void *pp2));
static int		cmp_size	ARGS ((const void *pp1, const void *pp2));
static int		cmp_type	ARGS ((const void *pp1, const void *pp2));
static int		cmp_uid		ARGS ((const void *pp1, const void *pp2));
static int		cmp_uid_number	ARGS ((const void *pp1, const void *pp2));
static int		cmp_write	ARGS ((const void *pp1, const void *pp2));
static int 		(*cmp_fcn)	ARGS ((const void *pp1, const void *pp2));

const char sortstr[] = "cgGilnNrstuUw";	/* must be in order CREATE..WRITE  */

	/* NB: DEC ULTRIX 4.3 /bin/cc compiler bug: cannot use pp1 and pp2 here: */
static int (*cmp_table[]) ARGS ((const void * /* pp1 */, const void * /* pp2 */)) =
{	/* NB: MUST be in order CREATE..WRITE, matching sortstr[] */
    cmp_create,
    cmp_gid,
    cmp_gid_number,
    cmp_inode,
    cmp_links,
    cmp_name,
    cmp_number,
    cmp_read,
    cmp_size,
    cmp_type,
    cmp_uid,
    cmp_uid_number,
    cmp_write,

    cmp_name,	/* final entry in case NUL is matched in set_sorttyp() */
};

static int		cmp_secondary ARGS ((const void *pp1, const void *pp2));

#if STDC
int
main (int argc, char *argv[])
#else /* NOT STDC */
int
main (argc, argv)
int argc;
char *argv[];
#endif /* STDC */
{
    int arg_error_count;
    char **oldargv;
    char *cp;
    int diredstat;
    int mark_stack[MAXMARKS];
    int mark_top;
    int special;		/* flag says had % or # in cmd */
    register char *t;
    register int cc;
    register int i;
    static char combuf[MAXFILEPATHLEN] = " ";		/* Holds last ! command */
    wait_arg_t status;

    prog = (argv[0] == (char *) NULL) ? PACKAGE_NAME : argv[0];
    t = strrchr(prog,'/');
    if (t != (char*)NULL)
	prog = t + 1;			/* reduce prog to basename of pathname */

    setbuf (stdout, bufout);	/* turn on buffering for faster output */

#if defined(HAVE_GRP_H)
    have_gid = False;
#endif

#if defined(HAVE_PWD_H)
    have_uid = False;
#endif

    batch_mode = isatty (fileno (stdout)) ? False : True;

    if (batch_mode)
	scrlen = INT_MAX;

    cmp_fcn = cmp_table[NAME];	/* initialize default sort comparison function */

    arg_error_count = 0;
    mark_top = 0;
    view_top = 0;
    re_comp ("/");		/* create an unmatchable initial file pattern */

    /* Get terminal type */

    Getcap ();
    Tscrlen--;			/* don't use final line, in case it is terminal status
				   line (e.g. line 25 on VT100 display) */

    /* Process arg flags. They must be first! */
    for (i = 1; i < argc; i++)
    {
	char *the_arg;

	if (argv[i][0] != '-')
	    break;
	the_arg = argv[i];
	if (the_arg[1] == '-')		/* allow GNU-style --arg */
	    the_arg++;
	switch ((cc = (int)the_arg[1], cc))
	{
	case 'b':
	    batch_mode = True;
	    scrlen = INT_MAX;
	    break;

	case 'd':
	    camedown = True;
	    break;

	case 'F':
	case 'f':		/* get the start file */
	    (void) strcpy (startfile, &the_arg[2]);
	    wasstartfile = True;
	    if (cc == 'F')
		quitok = False;
	    break;

	case 'g':
	    {
#if defined(HAVE_GRP_H)
		/* new option for version 4.00 [NHFB] */
		char *p;

		p = (char *) &the_arg[2];
		if (*p == '\0')	/* -g means current group id */
		    Gid = getgid ();
		else
		    /* expect -ggid or -ggroupname */
		{
		    struct group *g;

		    g = isdigit (*p) ? getgrgid ((gid_t) atoi (p)) : getgrnam (p);
		    if (g == (struct group *) NULL)
		    {
			FPRINTF (stderr, "Cannot map %s to valid group id\n",
				 argv[i]);
			my_exit (EXIT_FAILURE);
		    }
		    Gid = g->gr_gid;
		}
		have_gid = True;
#else
		FPRINTF (stderr, "%s option is not supported in this version\n",
			 argv[i]);
#endif
	    }
	    break;

	case '?':
	case 'h':
	    help_and_exit();
	    break;

	case 'm':
	    monochrome = True;
	    break;

	case 'S':			/* new option for dired-4.05 */
	    show_link_target = False;
	    break;

	case 'r':
	case 's':		/* Initial sort order */
	    set_sorttyp(cc,(int)the_arg[2]);
	    break;

	case 'u':
	    {
#if defined(HAVE_PWD_H)
		/* new option for version 4.00 [NHFB] */
		char *p;

		p = (char *) &the_arg[2];
		if (*p == '\0')	/* -u means current user id */
		    Uid = getuid ();
		else
		    /* expect -uuid or -uusername */
		{
		    struct passwd *u;

		    u = isdigit (*p) ? getpwuid ((uid_t) atoi (p)) : getpwnam (p);
		    if (u == (struct passwd *) NULL)
		    {
			FPRINTF (stderr, "Cannot map %s to valid user id\n",
				 argv[i]);
			my_exit (EXIT_FAILURE);
		    }
		    Uid = u->pw_uid;
		}
		have_uid = True;
#else
		FPRINTF (stderr, "%s option is not supported in this version\n",
			 argv[i]);
#endif
	    }
	    break;

	case 'V':		/* program version */
	case 'v':		/* program version */
	    {
		FPRINTF (stderr, "%s version %s %s\n",
			 prog, PACKAGE_VERSION, PACKAGE_DATE);
		my_exit (EXIT_SUCCESS);
	    }
	    break;

	case 'w':		/* Window size */
	    if (the_arg[2] == 'h')
		scrlen = 999;	/* Half */
	    else if (the_arg[2] == 'f')
		scrlen = 0;	/* Full */
	    else
	    {
		scrlen = atoi (&the_arg[2]);
		asswin = scrlen;
	    }
	    break;

	default:
	    FPRINTF (stderr, "ERROR: unknown option %s.\n", argv[i]);
	    arg_error_count++;
	    break;

	}
    }
    if (arg_error_count > 0)
	my_exit (EXIT_FAILURE);

    argc -= (i - 1);
    oldargv = argv;
    argv += (i - 1);

    windsize ();		/* replaced by proc call 8/82 crh */

    SIGNAL (SIGINT, SIG_IGN);
    SIGNAL (SIGQUIT, SIG_IGN);

    setdpy ();
    if (!batch_mode)
    {
	PRINTF ("Reading");
	(void) fflush (stdout);
    }
    lflg = True;
    iflg = sflg = False;
    numarg = argc;
    getfiles (argc, argv);

    /* extract the full path name */
    if (!arglist)
    {
	if (argc == 1)
	    cp = getfullpath (".");
	else if (argv[1][0] != '/')
	    cp = getfullpath (argv[1]);
	else
	    cp = argv[1];
	dcanon (cp);
	(void) strcpy (fullpath, cp);
    }

    qsort (file, totfiles, sizeof (struct lbuf), cmp_fcn);
    if (wasstartfile)
	locstart (startfile);
    else
	dispwin (0);

  restart:

    on_signals ();

    if (!arglist && !batch_mode)
	printpath (fullpath, True);
    countarg = 0;
  startup:
    while (!batch_mode)
    {
	command = (updateWinsize ? CTRL ('l') : getchar ());
	if (command == 'q')
	    break;		/* here's the loop exit */

#if defined(_AIX)
	showscreen ();		/* Workaround for curses library bug in older  */
	/* AIX versions on IBM systems; it is harmless */
	/* to leave it here even after the bug is fixed. */
#endif /* _AIX */

	if (blurb)
	{
	    telluser ("", "");
	    blurb = 0;
	}
	if (isdigit (command))
	    getcountarg ();

	/* GBG fix for cursor keys on DEC and Sun SPARC keyboards */
	/* Change dired.hlp to reflect new keys */
	if ((command == ESC) &&
	    (((i = getchar ()) == ESC) || (i == 'O') || (i == '[')))
	{
	    switch (getchar ())
	    {
	    case ESC:
		break;		/* Two escapes mapped to one */

	    case 'x':		/* fall through */
	    case 'A':
		command = 'k';
		break;		/* up arrow */

	    case 'r':		/* fall through */
	    case 'B':
		command = 'j';
		break;		/* down arrow */

	    case 'v':		/* fall through */
	    case 'C':
		command = 'f';
		break;		/* right arrow */

	    case 't':		/* fall through */
	    case 'D':
		command = 'b';
		break;		/* left arrow */

	    case 's':		/* fall through */
	    case 'k':
		command = 'f';
		break;		/* PageDown */

	    case 'y':
		command = 'b';
		break;		/* PageUp */

	    case '5':		/* ESC [ 5 ~ (Sun type 5) = PageUp */
		command = 'b';
		(void) getchar ();
		break;

	    case '6':		/* ESC [ 6 ~ (Sun type 5) = PageUp */
		command = 'f';
		(void) getchar ();
		break;
	    }
	}

	switch (command)
	{
	case ESC:
	    break;

	case ' ':
	    if (splitflg)
		prtsplitstr ("");	/* clear lower screen */
	    if (!arglist)
		printpath (fullpath, True);
	    break;

	case 'a':		/* Abort completely */
	    curxy (0, Tscrlen);
	    ceol (0, Tscrlen);
	    unsetdpy ();
	    my_exit (EXIT_FAILURE);
	    break;

	case 'Q':		/* abort all the direds and pushd to the
				   dir of the current file */
	    {
		char bldbuf[MAXFILEPATHLEN];

		bldbuf[0] = '\0';
		bldnam (bldbuf, numarg, curfile, argv);
		cp = strrchr (bldbuf, '/');
		curxy (0, Tscrlen);
		ceol (0, Tscrlen);
		unsetdpy ();
		if (cp)
		{
		    *cp = '\0';
		    typein ("pushd ");
		    typein (bldbuf);
		    typein ("\n");
		}
		my_exit (ABORTSTAT);
	    }
	    /* NOTREACHED */
	    break;

	case 'A':		/* Abort the chain of direds  ***crh */
	    curxy (0, Tscrlen);
	    ceol (0, Tscrlen);
	    unsetdpy ();
	    my_exit (ABORTSTAT);
	    break;

	case 'P':		/* Print a file */
	    if (file[curfile].ltype == 'd')
	    {
		telluser ("?Can only print files", "");
		break;

	    }
	    tmpbuf[0] = '\0';
	    bldnam (tmpbuf, numarg, curfile, argv);
	    telluser ("Printing...", "");
	    while ((i = (int) fork ()) == -1)
		sleep (3);
	    if (i == 0)
	    {
		execlp ("print", "print", tmpbuf, 0);
		telluser ("?Can't find program to list file.\n", "");
		my_exit (EXIT_FAILURE);
	    }
	    wait (&status);
	    break;

	case 'p':
	    {
		char bldbuf[MAXFILEPATHLEN];
		(void) strcpy (bldbuf, FNAME (curfile));
		expandpath (bldbuf);
		printpath (bldbuf, False);
	    }
	    break;

	case '/':
	    {
		char buff[MAXFILEPATHLEN];
		static char last_buff[MAXFILEPATHLEN];

		telluser (" ", "");
		curxy (0, Tscrlen);
		putchar ('/');
		if (!getline (buff))
		{
		    telluser ("?Pattern too long", "");
		    break;

		}
		if (buff[0] == '\0')	/* then re-use last search string */
		    (void) strcpy (buff, last_buff);
		if (((int) strlen (buff) > 0) && (re_comp (buff) != (char *) NULL))
		{
		    telluser ("?Pattern is not a valid regular expression", "");
		    break;

		}
		(void) strcpy (last_buff, buff);	/* save search string */
		findfile (1);
		break;

	    }

	case '\\':		/* toggle split screen */

	    if (splitflg)
		scrlen = Tscrlen - 1;
	    else
	    {
		if (asswin != -1)
		    scrlen = asswin;
		else
		    scrlen = 999;
	    }
	    windsize ();
	    topfile = curfile / (scrlen - 1) * (scrlen - 1);
	    blank ();
	    showscreen ();
	    curline = curfile - topfile;
	    curxy (CURSOR, curline);
	    break;

	case 'n':
	    findfile (1);
	    break;

	case 'N':
	    findfile (0);
	    break;

	case '!':		/* Execute a system command */
	case '.':		/* repeat the previous ! command */
	    telluser ("", "");
	    curxy (0, Tscrlen);
	    unsetdpy ();
	    PRINTF ("Command: ");
	    (void) fflush (stdout);
	    tmpbuf[0] = 'x';	/* dummy kludge */
	    if ((command == '!' ? fgets (tmpbuf, (int)sizeof(tmpbuf), stdin) :
		 strcpy (tmpbuf, combuf))
		!= NULL && tmpbuf[0] != '\0')
	    {
		register char *op,	/* old ptr */
		   *np;		/* new ptr */
		char bldbuf[MAXFILEPATHLEN];

		op = strchr(tmpbuf,(int)'\n'); /* eliminate any final newline char */
		if (op != (char*)NULL)
		    *op = '\0';

		(void) strcpy (combuf, tmpbuf);		/* remember the command */
		bldbuf[0] = '\0';
		op = tmpbuf;
		special = 0;
		while ((cc = *(np = skipto (op, "%#"))) != 0)
		{
		    special++;	/* set flag */
		    *np++ = '\0';	/* zap the %/# and bump past it */
		    (void) strcat (bldbuf, op);		/* 1st part */
		    if (cc == '%')	/* complete file name */
			bldnam (bldbuf, numarg, curfile, argv);
		    else
		    {		/* Had # sign, trailing comp only */
			if (numarg <= 2)
			    (void) strcat (bldbuf, file[curfile].ln.lname);
			else
			    (void) strcat (bldbuf, file[curfile].ln.namep);
		    }

		    op = np;
		}
		strcat (bldbuf, op);

		blank ();
		if (special)
		{		/* display expanded command */
		    PRINTF ("%s\n", bldbuf);
		    (void) fflush (stdout);
		}		/* "system" takes long enuf for him to see it */
		off_signals ();
		cshsystem (bldbuf);
		on_signals ();
		SIGNAL (SIGINT, SIG_IGN);
		SIGNAL (SIGQUIT, SIG_IGN);
		PRINTF ("\nCR to return...");
		(void) fflush (stdout);
		setdpy ();
		getchar ();
		blank ();
		showscreen ();
		telluser ("", "");
	    }
	    else
	    {			/* CR only, or EOF, or error */
		/*
		 * He changed his mind, skip it.  Since we were in
		 * cooked mode, the CR ending the gets is echoed and
		 * we lost the first line of the display.  So we
		 * re-insert it.
		 */
		setdpy ();
		if (tmpbuf[0] == 'x')	/* means EOF */
		    telluser ("", "");
		/* null entry, normal case */
		else if (my_AL == 0)
		{		/* no insert line capability */
		    blank ();
		    showscreen ();
		}
		else
		{		/* be a little sneakier */
		    curxy (0, Tscrlen - 1);	/* go back to where prompt */
		    ceol (0, Tscrlen - 1);	/*  is now, and blank it */
		    home ();
		    insline ();	/* make some room */
		    pentry (topfile);
		    putchar ('\n');
		}
		curxy (CURSOR, curline);
	    }
	    break;

	case 'g':		/* re-read the directory */
	case 'R':		/* re-read the directory */
	case 'X':		/* re-read the directory */
	    {
		int num;

		/* remember the name of the current file */
		strcpy (tmpbuf, FNAME (curfile));
		deletefiles (argc, argv);	/* delete marked files */
		if (numdeleted)
		    goto startup;	/* undeleted - restart */
		totfiles = 0;
		SIGNAL (SIGTSTP, SIG_DFL);
		curxy (0, Tscrlen);
		ceol (0, Tscrlen);
		if (!batch_mode)
		{
		    PRINTF ("Reading");
		    (void) fflush (stdout);
		}
		getfiles (argc, argv);
		qsort (file, totfiles, sizeof (struct lbuf), cmp_fcn);
		/* try to pos cursor where we were */
		if (findfname (tmpbuf, &num))
		    dispwin (num);
		else
		    dispwin (0);
		goto restart;
	    }


	case 'r':		/* Reverse sort */
	    curxy (0, Tscrlen);
	    PRINTF ("reverse ");
	    rflg = -1;
	    /* FALL THROUGH */

	case 's':		/* Normal sort */
	    if (command == 's')
	    {
		curxy (0, Tscrlen);
		rflg = 1;
	    }
	    PRINTF ("sort by [%s]: ",sortstr);
	    (void) fflush (stdout);
	    command = getchar ();
	    {
		int char_count;

		char_count = 0;
		while ((command == '?') || (strchr(sortstr,command) == (char*) NULL))
		{			/* offer more detailed help */
		    if (command == ESC)	/* cancel search request */
			break;
		    curxy (0, Tscrlen);
		    ceol (0, Tscrlen);
		    if (rflg == -1)
			PRINTF ("reverse ");
		    PRINTF ("sort: size, name, create/read/write, group, inode, links, type, user: ");
		    (void) fflush (stdout);
		    command = getchar ();
		    if (char_count++ > 0)
			telluser ("\007", "");
		}
	    }
	    if (command == ESC)		/* cancel search request */
	    {
		telluser ("", "");
		curxy (CURSOR, curline);
		break;

	    }

	    set_sorttyp((int)((rflg > 0) ? 's' : 'r'), command);
	    PRINTF ("%c", command);
	    (void) fflush (stdout);
	    qsort (file, totfiles, sizeof (struct lbuf), cmp_fcn);
	    topfile = 0;
	    curfile = 0;
	    curline = 0;
	    blank ();
	    showscreen ();
	    curxy (CURSOR, 0);
	    break;

	case 'S':		/* toggle showing of symbolic-link targets */
	    show_link_target = show_link_target ? False : True;
	    refresh_screen ();
	    break;

	case 'e':		/* Edit a file or directory */
	case 'E':		/* go up a dir */
	    if ((file[curfile].ltype == 'd') || (command == 'E'))
	    {
		strcpy (tmpbuf, dirednam);
		strcat (tmpbuf, " ");
		catargs (tmpbuf, oldargv);
	    }
	    else
	    {
		if (((t = getenv ("EDITOR")) != NULL) && ((int) strlen (t) > 0))
		    (void) strcat (strcpy (tmpbuf, t), " ");
		else
		    (void) strcat (strcpy(tmpbuf, DEFAULT_EDITOR), " ");
	    }
	    (void) strcpy (tmpbuf2, FNAME (curfile));
	    expandpath (tmpbuf2);
	    if (command == 'E')
	    {
		/* if we got where we are by going down, then 'E'
		   should just quit    */
		if (camedown)
		    goto quit;

		/* otherwise form the full path name of the upper node */
		cp = strrchr (tmpbuf2, '/');
		if (cp)
		{
		    *cp = '\0';
		    formstartarg (tmpbuf2);
		    (void) strcat (tmpbuf2, "/..");
		}
		else
		    break;	/* something would be wrong! */

		dcanon (tmpbuf2);
		if (STRING_EQUAL (tmpbuf2, fullpath))
		{
		    if (STRING_EQUAL (tmpbuf2, "/"))
			telluser ("?Already at '/'", "");
		    else
			telluser ("?Can't go up", "");
		    break;

		}
		strcat (tmpbuf, startarg);
	    }
	    else /* 'e' */
	    {
		if (file[curfile].ltype == 'd')
		{
		    if (wasstartfile && STRING_EQUAL (startfile, FNAME (curfile)) && quitok)
			/* we came up this path; just quit */
			goto quit;
		    (void) strcat (tmpbuf, "-d ");	/* indicate we're going down */
		}
	    }
	    (void) strcat (tmpbuf, tmpbuf2);
	    blank ();
	    unsetdpy ();
	    SIGNAL (SIGTSTP, SIG_DFL);
	    diredstat = system (tmpbuf);
	    if ((diredstat >> 8) == ABORTSTAT)
		my_exit (ABORTSTAT);
	    on_signals ();
	    setdpy ();
	    if (diredstat != 0)
	    {
		PRINTF ("\n type <cr> to continue");
		(void) fflush (stdout);
		getchar ();
	    }
	    blank ();
	    showscreen ();
	    if (arglist)
		telluser ("", "");
	    else
		printpath (fullpath, True);
	    break;

	case 'm':		/* 'more' a file */
	case 'v':		/* view a file (emacs dired binding) */
	    if (file[curfile].ltype == 'd')
	    {
		telluser ("?Can only page thru files", "");
		break;

	    }
	    {			/* allow PAGER environment variable to override builtin */
		char *p = getenv ("PAGER");
		strcpy (tmpbuf, p ? p : MOREPGM);
		strcat (tmpbuf, " ");	/* supply separating space */
	    }
	    bldnam (tmpbuf, numarg, curfile, argv);
	    blank ();
	    unsetdpy ();
	    SIGNAL (SIGTSTP, SIG_DFL);
	    system (tmpbuf);
	    on_signals ();
	    /* if (!sigint) {  */
	    PRINTF ("\nCR to return...");
	    (void) fflush (stdout);
	    /* } */
	    setdpy ();
	    getchar ();
	    blank ();
	    showscreen ();
	    telluser ("", "");
	    break;

	case 'T':		/* don't wait at page end */
	case 't':		/* quickly type the file -- added 5/81, J.Lepreau */
	    if (file[curfile].ltype == 'd')
	    {
		telluser ("?Can only type files", "");
		break;

	    }
	    tmpbuf[0] = '\0';
	    bldnam (tmpbuf, numarg, curfile, argv);
	    if (type (tmpbuf, (command == 't') ? True : False))
	    {			/* little t means wait */
		blank ();
		showscreen ();
		telluser ("", "");
	    }
	    curxy (CURSOR, curline);
	    break;

	case 'C':		/* toggle colorization */
	    monochrome = monochrome ? False : True;
	    refresh_screen();
	    break;

	case 'l':		/* Refresh screen */
	case CTRL ('l'):	/* added for editor compatibility -fjl */
	    refresh_screen();
	    break;

	case 'c':		/* Refresh current line */
	    curxy (0, curline);
	    pentry (curfile);
	    curxy (CURSOR, curline);
	    break;

	case CTRL ('v'):
	case 'f':		/* forward window */
	    fscreen (countarg);
	    break;

	    /* wish we could do meta-v */
	case 'b':		/* backward window */
	    bscreen (countarg);
	    break;

	case CTRL ('n'):
	case '\r':
	case 'j':		/***crh***/
	case '\n':		/* next file */
	    if (curfile == totfiles - 1)
		telluser ("?At end of files", "");
	    else
	    {
		if (countarg < 1)
		    countarg = 1;
		curfile = curfile + countarg;
		if (curfile > totfiles - 1)
		{
		    curfile = totfiles - 1;
		    dispfile (curfile);
		    shortcountmess ();
		    break;

		}
		dispfile (curfile);
	    }
	    break;

	case '^':		/* previous file */
	case CTRL ('h'):	/* backspace */
	case 'k':		/****crh****/
	case '-':
	case CTRL ('p'):
	    if (curfile == 0)
		telluser ("?At start of files", "");
	    else
	    {
		if (countarg < 1)
		    countarg = 1;
		curfile = curfile - countarg;
		if (curfile < 0)
		{
		    curfile = 0;
		    dispfile (curfile);
		    shortcountmess ();
		    break;

		}
		dispfile (curfile);
	    }
	    break;

	    /* ^G info as in vi  */
	case CTRL ('G'):
	    SPRINTF (tmpbuf, "file %d of %d --%d%%--",
		     curfile + 1, (int)totfiles, (curfile + 1) * 100 / (int)totfiles);
	    telluser ("%s", tmpbuf);
	    break;

	    /* line number addressing as in vi */
	case 'G':
	    if (countarg < 1)
		dispfile (totfiles - 1);
	    else if (countarg > totfiles)
	    {
		dispfile (totfiles - 1);
		shortcountmess ();
	    }
	    else
		dispfile (countarg - 1);
	    break;

	case 'h':		/* Help */
	case '?':
	    if (type (helpfile, True))
	    {			/* wait */
		blank ();
		showscreen ();
		telluser ("", "");
	    }
	    curxy (CURSOR, curline);
	    break;

	case 'd':		/* delete file */
	    if (file[curfile].deleted != DELETED)
	    {
		numdeleted++;
		file[curfile].deleted = DELETED;
		PRINTF ("D");
		(void) fflush (stdout);
	    }
	    if (curfile == totfiles - 1)
		curxy (CURSOR, curline); /***crh 1/84 ***/
	    else
	    {
		if (curline + 1 == scrlen)
		    fscreen (1);
		downline ();
	    }
	    break;

	case 'D':		/* delete all files */
	    for (i = 0; i < totfiles; ++i)
	    {
		if (file[i].deleted != DELETED)
		    file[i].deleted = DELETED;
	    }
	    numdeleted = totfiles;
	    blank ();
	    showscreen ();
	    telluser ("", "");
	    break;

	case '~':		/* delete editor backups of form "file~" */
	    for (i = 0; i < totfiles; ++i)
	    {
		size_t n = strlen (file[i].ln.lname);

		if ((file[i].ln.lname[n - 1] == '~') &&
		    (file[i].deleted != DELETED))
		{
		    numdeleted++;
		    file[i].deleted = DELETED;
		}
	    }
	    blank ();
	    showscreen ();
	    telluser ("", "");
	    break;

	case '#':		/* delete editor backups of form "#file" */
	    for (i = 0; i < totfiles; ++i)
	    {
		if ((file[i].ln.lname[0] == '#') &&
		    (file[i].deleted != DELETED))
		{
		    numdeleted++;
		    file[i].deleted = DELETED;
		}
	    }
	    blank ();
	    showscreen ();
	    telluser ("", "");
	    break;

	case 'U':		/* undelete all requested deletions */
	    for (i = 0; i < totfiles; ++i)
	    {
		if (file[i].deleted == DELETED)
		    file[i].deleted = NOTDELETED;
	    }
	    numdeleted = 0;
	    blank ();
	    showscreen ();
	    telluser ("", "");
	    break;

	case 'u':		/* undelete file */
	    if (file[curfile].deleted == DELETED)
	    {
		numdeleted--;
		file[curfile].deleted = NOTDELETED;
		PRINTF (" ");
		(void) fflush (stdout);
	    }
	    if (curfile == totfiles - 1)
		curxy (CURSOR, curline); /***crh 1/84 ***/
	    else
	    {
		if (curline + 1 == scrlen)
		    fscreen (1);
		downline ();
	    }
	    break;

	case '[':
	case '(':
	case '{':
	case '<':		/* go to start of file list */
	    if (countarg <= 0)
		dispfile (0);
	    else
	    {
		if (countarg > 10)
		    countarg = 10;
		dispfile ((int) (((long) countarg * (long) totfiles) / 10L));
	    }
	    break;

	case ']':
	case ')':
	case '}':
	case '>':		/* go to end of file list */
	    if (countarg <= 0)
		dispfile (totfiles - 1);
	    else
	    {
		if (countarg > 10)
		    countarg = 10;
		countarg = 10 - countarg;
		dispfile ((int) (((long) countarg * (long) totfiles) / 10L));
	    }
	    break;

	case 'J':		/* jump to top mark, pop mark stack */
	    if (mark_top > 0)
		dispfile (mark_stack[--mark_top]);
	    else
		telluser ("?Mark stack is empty: maybe push some with M", "");
	    break;

	case 'L':		/* jump to top view, pop view stack */
	    if (view_top > 0)
	    {
		dispfile (view_stack[--view_top]);
		view_top--;	/* undo the view_top++ in dispfile() */
	    }
	    else
		telluser ("?View stack is empty", "");
	    break;

	case 'M':		/* save mark */
	    if (mark_top < MAXMARKS)
	    {
		mark_stack[mark_top++] = curfile;
		telluser ("Marked...", "");
	    }
	    else
		telluser ("?Too many marks pushed: pop some with J", "");
	    break;

	default:
	    if (command != -1)	/* get around ^Z sending EOF? */
		telluser ("?Unknown command. Type ? or h for help ", "");
	    break;

	}
	countarg = 0;
    }
  quit:
    deletefiles (argc, argv);
    if (numdeleted)
	goto startup;
    curxy (0, Tscrlen);		/*** crh ***/
    ceol (0, Tscrlen);
    unsetdpy ();
    my_exit(errcode);
    /* NOTREACHED */
    return(errcode);
}

static void
getcountarg (VOID_ARG)
{
    countarg = 0;
    while (isdigit (command))
    {
	countarg = 10 * countarg + (command - (int) '0');
	command = getchar ();
    }
}


#if STDC
static void
typein (const char *argv)
#else /* NOT STDC */
static void
typein (argv)
const char *argv;
#endif /* STDC */
{				/* put arg chars into input buffer, i.e., fake typing them in */
#if defined(USE_SGTTY) && defined(TIOCGETP) && defined(TIOCSETN)
    struct sgttyb stb;
    struct sgttyb stb2;

    /*  int pendin = LPENDIN; */

    ioctl (2, TIOCGETP, &stb);
    stb2 = stb;
    stb.sg_flags &= ~ECHO;
    ioctl (2, TIOCSETN, &stb);
    for (;; argv++)
    {
	if (*argv == '\0')
	    break;
	/* else ioctl(2, TIOCSTI, argv); */
    }
    ioctl (2, TIOCSETN, &stb2);
    /* ioctl(2, TIOCLBIS, &pendin); */
#else
    FPRINTF (stderr, "typein(\"%s\") not yet implemented\n", argv);
#endif /* defined(USE_SGTTY) && defined(TIOCGETP) && defined(TIOCSETN) */
}


static void
off_signals (VOID_ARG)
{
    SIGNAL (SIGINT, SIG_DFL);
    SIGNAL (SIGQUIT, SIG_DFL);
    SIGNAL (SIGTSTP, SIG_DFL);
#if defined(SIGWINCH)
    SIGNAL (SIGWINCH, SIG_DFL);
#endif
}


#if STDC
static handler_t
onctlz (int sig)
#else /* NOT STDC */
static handler_t
onctlz (sig)
int sig;
#endif /* STDC */
{
    sig = sig;			/* avoid compiler unused-argument warnings */
    curxy (0, Tscrlen);
    ceol (0, Tscrlen);
    unsetdpy ();
    SIGNAL (SIGTSTP, SIG_DFL);
#if defined(SIGSTOP)
    kill (0, SIGSTOP);		/*re-send ^z */
#endif
    SIGNAL (SIGTSTP, onctlz);	/* wake up here ... */
    setdpy ();
    blank ();
    showscreen ();
    if (!intype)
	telluser ("", "");
    else
    {
	curxy (0, Tscrlen);
	ceol (0, Tscrlen);
    }
#if !defined(HAVE_VOID_HANDLER)
    return (0);
#endif
}


static void
on_signals (VOID_ARG)
{
    SIGNAL (SIGTSTP, onctlz);

#if defined(SIGWINCH)
    SIGNAL (SIGWINCH, win_change);
#endif
}


static void
refresh_screen (VOID_ARG)
{
	getWinSize ();	/* May be in X, new window size */
	windsize ();	/* New divider */
	blank ();
	showscreen ();
	telluser ("", "");
	updateWinsize = False;
}


#if defined(SIGWINCH)
#if STDC
static handler_t
win_change (int sig)
#else /* NOT STDC */
static handler_t
win_change (sig)
int sig;
#endif /* STDC */
{
    sig = sig;			/* avoid compiler unused-argument warnings */
    SIGNAL (SIGWINCH, SIG_IGN);
    updateWinsize = True;		/* schedule an update */
    SIGNAL (SIGWINCH, win_change);
#if !defined(HAVE_VOID_HANDLER)
    return (0);
#endif
}
#endif /* defined(SIGWINCH) */


#if STDC
static int
match (int filenum)
#else /* NOT STDC */
static int
match (filenum)
int filenum;
#endif /* STDC */
{
    char *strp;
    strp = (file[filenum].lflags & ISARG) ? file[filenum].ln.namep :
	file[filenum].ln.lname;
    return (re_exec (strp));
}


#if STDC
static void
dispfile (int fnum)
#else /* NOT STDC */
static void
dispfile (fnum)
int fnum;
#endif /* STDC */
{				/* move to the indicated file number */
    if (view_top < MAXVIEW)	/* remember the last view shown */
	view_stack[view_top++] = curfile;
    else
	telluser ("?Too many views pushed: pop some with L", "");

    /* redraw if file not already on screen */
    if ((fnum < topfile) || (fnum >= topfile + scrlen))
    {
	topfile = fnum / (scrlen - 1) * (scrlen - 1);

	/***crh 2/84 get rid of 1 file final windows unless necessary ***/

	if ((topfile == totfiles - 1) && (topfile - scrlen + 1 >= 0))
	    topfile = topfile - scrlen + 1;

	blank ();
	showscreen ();
    }
    curfile = fnum;
    curline = fnum - topfile;
    curxy (CURSOR, curline);
}


#if STDC
static void
findfile (int down)
#else /* NOT STDC */
static void
findfile (down)
int down;
#endif /* STDC */
{
    int i;
    int status;

    status = 0;
    if (down)
    {
	for (i = curfile + 1; i < totfiles; i++) /***crh 1/84 ch from <=*/
	{
	    status = match (i);
	    if (status != 0)
		break;
	}
    }
    else
    {
	for (i = curfile - 1; i >= 0; i--)
	{
	    status = match (i);
	    if (status != 0)
		break;
	}
    }
    if (status == -1)
    {
	telluser ("?Botched regular expression", "");
	return;
    }
    if (status)
	dispfile (i);
    else
	telluser ("?No match", "");
}


#if STDC
static int
getline (char s[])
#else /* NOT STDC */
static int
getline (s)
char s[];
#endif /* STDC */
{
    int status;
    int c;
    int i;

    i = 0;
    s[0] = '\0';
    status = 1;
    while (((c = getchar ()) != EOF) && (c != '\n') && (c != '\r'))
    {
	if (c == KILL_CHAR)
	{
	    i = 0;
	    curxy (1, Tscrlen);
	    ceol (1, Tscrlen);
	    continue;
	}
	if (c != ERASE_CHAR)
	{
	    if (i < scrwid - 1)
	    {
		s[i++] = c;
		putchar (c);
	    }
	    else
		status = 0;
	}
	else
	{
	    if (i > 0)
	    {
		i--;
		putchar (c);
		putchar (' ');
		putchar (c);
	    }
	}
    }
    s[i] = '\0';
    return (status);
}


static void
windsize (VOID_ARG)
{
    register char *t;
    register char *tt;

    if (scrlen == 0)		/* full screen */
	scrlen = Tscrlen;
    else if (scrlen == 999)	/* means split in half */
	scrlen = (Tscrlen - 1) >> 1;
    if (scrlen < 2)
	scrlen = 2;
    if (Tscrlen < scrlen)
	scrlen = Tscrlen;
    splitflg = (Tscrlen > scrlen + 1);	/* 1 extra line for separator */
    if (splitflg)
    {
#if 0
	Worklen = Tscrlen - (scrlen + 1);	/* size of 'working' window */
#endif
	Worktop = scrlen + 1;	/* bottom half for now */
    }
#if 0
    else
	Worklen = 0;
#endif

    tt = divider + scrwid;
    LW = scrwid;		/* last position in the linear window */
    for (t = divider; t < tt;)	/* arbitrary length */
	*t++ = DIVCHAR;
    *t = '\0';
}


#if STDC
static void
locstart (char strtfile[])
#else /* NOT STDC */
static void
locstart (strtfile)
char strtfile[];
#endif /* STDC */
{
    int sfilenum;

    if (findfname (strtfile, &sfilenum))
	dispwin (sfilenum);
    else
    {
	/* if we cant find it, we cant use it */
	wasstartfile = False;
	dispwin (0);
    }

}


#if STDC
static Boolean
findfname (char filename[], int *num)
#else /* NOT STDC */
static Boolean
findfname (filename, num)
char filename[];
int *num;
#endif /* STDC */
{
    int i;

    for (i = 0; i < totfiles; i++) /***crh 1/84 ch from <=*/
    {
	if (STRING_EQUAL (FNAME (i), filename))
	{
	    *num = i;
	    return (True);
	}
    }
    return (False);
}


#if STDC
static void
formstartarg (char *name)
#else /* NOT STDC */
static void
formstartarg (name)
char *name;
#endif /* STDC */
{				/* form start file arg */
    char tmpbuf3[MAXFILEPATHLEN];
    char *cp;

    if (arglist)
	(void) strcpy (startarg, " -F");
    else
	(void) strcpy (startarg, " -f");
    (void) strcpy (tmpbuf3, name);
    dcanon (tmpbuf3);
    cp = strrchr (tmpbuf3, '/');
    if (cp)
	(void) strcat (startarg, cp + 1);
    else
	(void) strcat (startarg, tmpbuf3);
    (void) strcat (startarg, " ");
}


#if STDC
static void
dcanon (char *cp)
#else /* NOT STDC */
static void
dcanon (cp)
char *cp;
#endif /* STDC */
{	/* canonicalize the pathname, removing excess ./ and ../ etc.  */
    register char *p;
    register char *sp;
    register int slash;

    /* assume name begins with a '/' */

    for (p = cp; *p;)
    {				/* for each component */
	sp = p;			/* save slash address */
	while (*++p == '/')	/* flush extra slashes */
	    ;
	if (p != ++sp)
	    strcpy (sp, p);
	p = sp;			/* save start of component */
	slash = 0;
	while (*++p)		/* find next slash or end of path */
	    if (*p == '/')
	    {
		slash = 1;
		*p = 0;
		break;
	    }
	if (*sp == '\0')	/* if component is null */
	    if (--sp == cp)	/* if path is one char (i.e. /) */
		break;
	    else
		*sp = '\0';
	else if (STRING_EQUAL (".", sp))
	{
	    if (slash)
	    {
		strcpy (sp, ++p);
		p = --sp;
	    }
	    else if (--sp != cp)
		*sp = '\0';
	}
	else if (STRING_EQUAL ("..", sp))
	{
	    if (--sp != cp)
		while (*--sp != '/')
		    ;
	    if (slash)
	    {
		strcpy (++sp, ++p);
		p = --sp;
	    }
	    else if (cp == sp)
		*++sp = '\0';
	    else
		*sp = '\0';
	}
	else if (slash)
	    *p = '/';
    }
}


#if STDC
static char *
getfullpath (const char *start)
#else /* NOT STDC */
static char *
getfullpath (start)
const char *start;
#endif /* STDC */
{
    /*  return the full path name */

    static char temp[500];	/* address returned */

#ifndef V4P2
    char dot[500];
    char dotdot[500];
    int file;
    struct stat d;
    struct stat dd;
    struct dirent dir;
    int rdev;
    int rino;

    off = -1;
    (void) strcpy (dot, start);
    strcpy (dotdot, start);
    strcat (dotdot, "/..");
    stat ("/", &d);
    rdev = (int)d.st_dev;
    rino = (int)d.st_ino;
    for (;;)
    {
	stat (dot, &d);
	if (d.st_ino == rino && d.st_dev == rdev)
	{
	    prname ();
	    return (wd);
	}
	if ((file = open (dotdot, 0)) < 0)
	{
	    telluser ("?pwd: cannot open ..\n", "");
	    my_exit (EXIT_FAILURE);
	}
	fstat (file, &dd);
	strcat (dotdot, "/..");
	strcat (dot, "/..");
	if (d.st_dev == dd.st_dev)
	{
	    if (d.st_ino == dd.st_ino)
	    {
		prname ();
		return (wd);
	    }
	    do
		if (read (file, (char *) &dir, sizeof (dir)) < sizeof (dir))
		{
		    telluser ("?Read error in ..\n", "");
		    PRINTF (dir.d_name);
		    my_exit (EXIT_FAILURE);
		}
	    while (dir.d_ino != d.st_ino);
	}
	else
	    do
	    {
		if (read (file, (char *) &dir, sizeof (dir)) < sizeof (dir))
		{
		    telluser ("?read error in ..\n", "");
		    PRINTF (dir.d_name);
		    my_exit (EXIT_FAILURE);
		}
		strcpy (temp, dot);
		strcat (temp, "/");
		strcat (temp, dir.d_name);
		stat (temp, &dd);
	    }
	    while (dd.st_ino != d.st_ino || dd.st_dev != d.st_dev);
	close (file);
	{
	    register int i,
		j;

	    i = -1;
	    while (dir.d_name[++i] != 0);
	    if ((off + i + 2) > BUFSIZ - 1)
	    {
		prname ();
		return (wd);
	    }
	    for (j = off + 1; j >= 0; --j)
		name[j + i + 1] = name[j];
	    off = i + off + 1;
	    name[i] = '/';
	    for (--i; i >= 0; --i)
		name[i] = dir.d_name[i];
	}
    }
#else
    /* Working dir fix by GBG for local unix systems */
    if (getenv ("PWD") != (char *) 0)
	(void) strcpy (wd, getenv ("PWD"));
    else if (getenv ("cwd") != (char *) 0)
	(void) strcpy (wd, getenv ("cwd"));
    else if (!getwd (wd))
    {
	telluser (wd, "");
	my_exit (EXIT_FAILURE);
    }
    if (STRING_EQUAL (start, "."))
	return (wd);
    else
    {
	(void) strcpy (temp, wd);
	(void) strcat (temp, "/");
	(void) strcat (temp, start);
	return (temp);
    }
#endif
}


#ifndef V4P2
static void
prname (VOID_ARG)
{
    strcpy (wd, "/");
    if (off < 0)
	off = 0;
    name[off] = '\0';
    strcat (wd, name);
}
#endif


#if STDC
static void
prtsplitstr (const char str[])
#else /* NOT STDC */
static void
prtsplitstr (str)
const char str[];
#endif /* STDC */
{
    int i;
    int top;
    int cur_scrl;

    cur_scrl = totfiles - topfile + 1;	/* topfile starts at 0 */
    top = MIN(cur_scrl,scrlen) + 1;
    curxy (0, top);
    if (my_CD != 0)
	putpad (my_CD);
    for (i = 0; i < (int) strlen (str); i++)
    {				/*** 1/84 crh - show embedded control chars - assume ASCII ***/
	if (str[i] >= ' ')
	{
	    if (str[i] >= '\177')
	    {
		putchar ('^');
		putchar ('?');
	    }
	    else
		putchar (str[i]);
	}
	else
	{
	    putchar ('^');
	    putchar (str[i] + '@');
	}
    }
    curxy (CURSOR, curline);
}


#if STDC
static void
printpath (char str[], int printbottom)
#else /* NOT STDC */
static void
printpath (str, printbottom)
char str[];
int printbottom;
#endif /* STDC */
{
    static const char *error = "path too long - use split screen to display it";

    if (printbottom)
    {
	if ((int) strlen (str) < scrwid)
	    telluser (str, "");
	else if (splitflg)
	    prtsplitstr (str);
	else
	    telluser (error, "");
    }
    else if (splitflg)
	prtsplitstr (str);
    else if ((int) strlen (str) < scrwid)
	telluser (str, "");
    else
	telluser (error, "");
}


#if STDC
static void
expandpath (char pname[])
#else /* NOT STDC */
static void
expandpath (pname)
char pname[];
#endif /* STDC */
{
    char temp[MAXFILEPATHLEN];

    if (pname[0] != '/')
    {
	if (arglist)
	    strcpy (temp, getfullpath ("."));
	else
	    strcpy (temp, fullpath);
	strcat (temp, "/");
	strcat (temp, pname);
	strcpy (pname, temp);
    }
    dcanon (pname);
}


#if STDC
static void
getfiles (int argc, char *argv[])
#else /* NOT STDC */
static void
getfiles (argc, argv)
int argc;
char *argv[];
#endif /* STDC */
{
    if (argc == 1)
	Readdir (".");
    else
    {
	struct stat stbuf;
	arglist = 1;
	if (argc == 2)
	{
	    if (stat (argv[1], &stbuf) == -1)
	    {
		PRINTF ("\nSorry, %s unreadable.\n", argv[1]);
		unsetdpy ();
		my_exit (EXIT_FAILURE);
	    }
	    if ((stbuf.st_mode & S_IFMT) == S_IFDIR)
	    {
		Readdir (argv[1]);
		arglist = 0;
	    }
	}
	if (arglist)
	{
	    while (--argc > 0)
	    {
		if (totfiles == maxtotfiles)
		    grow_file_array();
		if (!batch_mode && ((totfiles % DIRPBLOCK) == 0))
		{
		    putchar ('.');
		    (void) fflush (stdout);
		}
		if (gstat (*++argv) == 0)
		{
		    file[totfiles].ln.namep = *argv;
		    file[totfiles].lflags |= ISARG;
		    totfiles++;
		}
	    }
	    getuids();
	    getgids();
	}
    }
    if (totfiles == 0)
    {
	PRINTF ("\n?Empty directory%s\007\n",
		((have_gid || have_uid) ? " or no matching files" : ""));
	unsetdpy ();
	sleep (1);		/* So user can see it - don't worry if less */
	my_exit (EXIT_SUCCESS);
    }

}


#if STDC
static void
deletefiles (int argc, char *argv[])
#else /* NOT STDC */
static void
deletefiles (argc, argv)
int argc;
char *argv[];
#endif /* STDC */
{
    int i;

    argc = argc;		/* avoid compiler unused-argument warnings */

    if (numdeleted)
    {
	blank ();
	PRINTF ("The following %s marked for deletion:\n",
		(numdeleted == 1) ? "is" : "are");
	(void) fflush (stdout);
	typefiles ();
	(void) fflush (stdout);
	PRINTF ("\nShall I delete %s? ",
		(numdeleted == 1) ? "this" : "these");
	(void) fflush (stdout);
	if ((command = getchar ()) != 'y')
	{
	    blank ();
	    showscreen ();
	    curxy (0, Tscrlen);
	    ceol (0, Tscrlen);
	    curxy (CURSOR, curline);
	    return;
	}
	else
	{
	    int failures;

	    failures = 0;
	    PRINTF ("y\n");
	    (void) fflush (stdout);
	    for (i = 0; i < totfiles; i++)
		if (file[i].deleted == DELETED)
		{
		    nambuf[0] = '\0';
		    bldnam (nambuf, numarg, i, argv);
		    if (file[i].ltype == 'd')
			rm (nambuf);
		    else if (unlink (nambuf) < 0)
		    {
			PRINTF ("Delete of %s failed.\n", nambuf);
			(void) fflush (stdout);
			failures++;
		    }
		}
	    if (failures)
	    {
		PRINTF ("<CR> to continue");
		(void) fflush (stdout);
		getchar ();
	    }
	}
    }
    numdeleted = 0;
}


static void
typefiles (VOID_ARG)
{
    int i;
    int j;
    int longsiz;
    int longthis;
    int maxperln;
    int numout;
    register char *cp;

    longsiz = numout = 0;
    for (i = 0; i < totfiles; i++)
	if (file[i].deleted == DELETED)
	    if (file[i].lflags & ISARG)
	    {
		if ((int) strlen (file[i].ln.namep) > longsiz)
		{
		    longsiz = strlen (file[i].ln.namep);
		}
	    }
	    else
	    {
		if ((int) strlen (file[i].ln.lname) > longsiz)
		{
		    longsiz = strlen (file[i].ln.lname);
		}
	    }
    maxperln = scrwid / (longsiz + 3);
    if (maxperln < 1)
	maxperln = 1;		/* 3/84 */
    for (i = 0; i < totfiles; i++)
	if (file[i].deleted == DELETED)
	{
	    cp = FNAME (i);
	    longthis = strlen (cp);
	    for (; *cp; cp++)
	    {
		if ((*cp < ' ') || (*cp == '\177'))
		    putchar ('?');
		else
		    putchar (*cp);
	    }
	    numout++;
	    if ((numout % maxperln) == 0)
		putchar ('\n');
	    else if (numout != numdeleted)
		for (j = 0; j < (longsiz + 3 - longthis); j++)
		    putchar (' ');
	}
}


#if STDC
static void
rm (char arg[])
#else /* NOT STDC */
static void
rm (arg)
char arg[];
#endif /* STDC */
{
#ifndef V4P2
    struct stat buf;
    struct dirent dirent;
    char name[100];
    int d;

    if (stat (arg, &buf))
    {
	return;
    }
    if ((buf.st_mode & S_IFMT) == S_IFDIR)
    {
	if (access (arg, 02) < 0)
	{
	    PRINTF ("%s not deleted.\n", arg);
	    (void) fflush (stdout);
	    return;
	}
	if ((d = open (arg, 0)) < 0)
	{
	    PRINTF ("rm: %s: cannot read\n", arg);
	    (void) fflush (stdout);
	    return;
	}
	while (read (d, (char *) &dirent, sizeof (dirent)) == sizeof (dirent))
	{
	    if (dirent.d_ino != 0 && !dotname (dirent.d_name))
	    {
		SPRINTF (name, "%s/%.14s", arg, dirent.d_name);
		rm (name);
	    }
	}
	close (d);
	errcode += Rmdir (arg);
	return;
    }

    if (unlink (arg))
    {
	++errcode;
	PRINTF ("%s not deleted.\n", arg);
	(void) fflush (stdout);
    }
#else
    (void) strcpy (tmpbuf, "rm -r ");
    (void) strcat (tmpbuf, arg);
    if (system (tmpbuf) >> 8)
	telluser ("?Couldn't delete %s", arg);
    return;
#endif
}


#ifndef V4P2
#if STDC
static int
dotname (char *s)
#else /* NOT STDC */
static int
dotname (s)
char *s;
#endif /* STDC */
{
    if (s[0] == '.')
	if (s[1] == '.')
	    if (s[2] == '\0')
		return (1);
	    else
		return (0);
	else if (s[1] == '\0')
	    return (1);
    return (0);
}


#if STDC
static int
Rmdir (char *f)
#else /* NOT STDC */
static int
Rmdir (f)
char *f;
#endif /* STDC */
{
    int i;
    wait_arg_t status;

    if (dotname (f))
	return (0);
    while ((i = fork ()) == -1)
	sleep (3);
    if (i == 0)
    {
	execl ("/bin/rmdir", "rmdir", f, 0);
	execl ("/usr/bin/rmdir", "rmdir", f, 0);
	PRINTF ("rm: can't find rmdir\n");
	my_exit (EXIT_FAILURE);
    }
    wait (&status);
    return (status);
}
#endif /* V4P2 */


static void
shortcountmess (VOID_ARG)
{
    telluser ("?Used shortened count", "");
}


#if STDC
static void
dispwin (int filenum)
#else /* NOT STDC */
static void
dispwin (filenum)
int filenum;
#endif /* STDC */
{				/* force a new window */
    topfile = -999;		/* force a redraw */
    dispfile (filenum);
}


#if STDC
static void
fscreen (int count)
#else /* NOT STDC */
static void
fscreen (count)
int count;
#endif /* STDC */
{
    if (topfile + scrlen - 1 >= totfiles - 1) /***crh 1/84 ch from > */
    {
	telluser ("?No remaining windows", "");
	return;
    }
    if (count <= 1)
	count = 1;
    else if ((topfile + count * (scrlen - 1)) > totfiles - 1)
    {
	dispwin (totfiles - 1);
	shortcountmess ();
	return;
    }
    dispwin (topfile + count * (scrlen - 1));
}


#if STDC
static void
bscreen (int count)
#else /* NOT STDC */
static void
bscreen (count)
int count;
#endif /* STDC */
{
    if (topfile - scrlen + 1 < 0)
    {
	telluser ("?No previous windows", "");
	return;
    }
    if (count <= 1)
	count = 1;
    else if ((topfile - count * (scrlen - 1)) < 0)
    {
	dispwin (0);
	shortcountmess ();
	return;
    }
    dispwin (topfile - count * (scrlen - 1));
}


static void
showscreen (VOID_ARG)
{
    int i;
    int numprint;
    int x1;
    int x2;

    home ();
    numprint = 0;
    for (i = topfile; (numprint < scrlen) && (i < totfiles); i++)
    {
	numprint++;
	pentry (i);
	putchar ('\n');
    }
    if (splitflg)
    {
	strcpy (newdivider, divider);
	x1 = (topfile * LW) / totfiles;
	x2 = ((topfile + scrlen - 1) * LW) / totfiles;
	if (x1 == x2)
	    newdivider[x1] = 'o';
	else
	{
	    if ((x1 == 0) && (topfile == 0))
		newdivider[0] = '[';
	    else
		newdivider[x1] = '(';
	    if (x2 >= LW || topfile + scrlen >= totfiles)
		newdivider[LW - 1] = ']';
	    else
		newdivider[x2 - 1] = ')';
	}
	PRINTF ("%s\n", newdivider);
    }
    (void) fflush (stdout);	/* reset for display functions */
}


#if STDC
static void
Readdir (const char *dir)
#else /* NOT STDC */
static void
Readdir (dir)			/* Reads directory dir */
const char *dir;
#endif /* STDC */
{
    struct dirent *dentryp;

#if defined(HAVE_OPENDIR)
    DIR *dirp;

    if ((dirp = opendir (dir)) == NULL)
#else
    FILE *dirf;

    if ((dirf = fopen (dir, "r")) == NULL)
#endif
    {
	PRINTF ("\nSorry, %s unreadable.\n", dir);
	(void) fflush (stdout);
	unsetdpy ();
	my_exit (EXIT_FAILURE);
    }
#if defined(HAVE_OPENDIR)
    while ((dentryp = readdir (dirp)) != NULL)
#else
    for (;;)
#endif
    {
#if !defined(HAVE_OPENDIR)
	struct dirent dentry;

	if (fread ((char *) &dentry, sizeof (dentry), 1, dirf) != 1)
	    break;
	dentryp = &dentry;
#endif
	if ((dentryp->d_ino == 0)
	    || ((dentryp->d_name[0] == '.') &&
		((dentryp->d_name[1] == '\0') ||
	    ((dentryp->d_name[1] == '.') && (dentryp->d_name[2] == '\0')))))
	    continue;

	if (totfiles == maxtotfiles)
	    grow_file_array();
	/* Just ignore if can't find the file, dir may be changing */
	if (gstat (makename (dir, dentryp->d_name, D_NAMLEN (dentryp))) == 0)
	{			/* 0 == Ok */
	    file[totfiles].lnum = dentryp->d_ino;
#if defined(HAVE_DIRENT_H)
	    (void) strcpy (file[totfiles].ln.lname, dentryp->d_name);
#else
	    {
		register int j;
		/* old <sys/dir.h> style might lack NUL termination */
		for (j = 0; j < D_NAMLEN (dentryp); j++)
		    file[totfiles].ln.lname[j] = dentryp->d_name[j];
		file[totfiles].ln.lname[j] = '\0';	/* ensure NUL termination */
	    }
#endif
	    totfiles++;
	    if (!batch_mode && (totfiles % 10 == 0))
		putchar ('.');
	}
    }
#if defined(HAVE_OPENDIR)
    closedir (dirp);
#else
    (void) fclose (dirf);
#endif
    getuids();
    getgids();
}


static void
getgids()
{
    /*******************************************************************
      Looking up alphanumeric equivalents of numeric group and user ids
      is potentially very expensive, since each lookup may involve
      client/server network communication.  We therefore optimize the
      operation by first sorting by the numeric identifier, then using
      getname(), knowing that it caches its results.  This ensures that
      no name is ever looked up more than once.

      This function, and the companion getuids() function, was
      introduced after finding that doing the name lookup in the inner
      loop of Readdir() was noticeably slow on some systems, compared to
      an earlier dired version that did not do group name lookups.

      A dramatic demonstration of the cost of name lookup is provided by
      the Silicon Graphics IRIX 5.3 pixie profiler, which produced the
      data in this table of the average clock cycles spent in a SINGLE
      call to each of various routines, sorted by descending cycle
      count; entries with fewer than 100 cycles/call have been
      discarded.  Notice also the significant difference in the cost of
      getgrgid() and getpwuid() (757 to 1):

      392293 _getgrgid            597 _doprnt              163 pmode__Fi
       28042 qsort                553 windsize__Fv         161 unescape__FPc
       10346 setupterm            521 tparm                157 dispfile__Fi
	9028 Readdir__FPCc        518 __sgi_getpwuid       144 _yp_match
	7230 qst                  406 get_cap__FPCc        135 interpret
	2624 inlist               341 clntudp_call         119 colorize__FPCcPC
	1836 offtime              268 dcanon__FPc          119 fflush
	1679 getgids__Fv          234 dirtok__FPcPCc       119 makekey
	1679 getuids__Fv          225 gstat__FPCc          116 interpret
	1180 main                 218 strtol               114 load_dom_binding
	1155 getenv               212 _tcsearch            113 fputs
	1077 parse_colors__FP     204 __start              109 _localtime_r
	 869 _flushlbf            195 __filbuf             106 _morecore
	 786 showscreen__Fv       183 Getcap__Fv           104 _memccpy
	 720 pentry__Fi           175 makename__FPCcT1     100 tputs
	 657 set_cap__FPCcPc      164 _clntudp_bufcrea

    *******************************************************************/

    size_t i;

    qsort (file, totfiles, sizeof (struct lbuf), cmp_gid);

    max_gname_width = 0;
    max_gid_width = 0;
    for (i = 0; i < totfiles; ++i)
    {
	int m;
	gid_t g;
	(void) getname (True, (long) file[i].lgid, file[i].groupname,
			sizeof (file[i].groupname));
	m = (int)strlen(file[i].groupname);
	max_gname_width = MAX(max_gname_width, m);
	m = 0;
	for (g = file[i].lgid, m = 0; g != 0U; g /= 10U, m++)
	    /* NO-OP */ ;
	max_gid_width = MAX(max_gid_width, m);
    }
}


static void
getuids()
{
    /* See getgids() above for detailed commentary on this function. */

    size_t i;

    qsort (file, totfiles, sizeof (struct lbuf), cmp_uid);

    max_lsize_width = 0;
    max_uname_width = 0;
    max_uid_width = 0;
    for (i = 0; i < totfiles; ++i)
    {
	int m;
	long n;
	uid_t u;
	(void) getname (False, (long) file[i].luid, file[i].username,
			sizeof (file[i].username));
	m = (int)strlen(file[i].username);
	max_uname_width = MAX(max_uname_width, m);
	for (u = file[i].luid, m = 0; u != 0U; u /= 10U, m++)
	    /* NO-OP */ ;
	max_uid_width = MAX(max_uid_width, m);
	for (n = file[i].lsize, m = 0; n != 0U; n /= 10U, m++)
	    /* NO-OP */ ;
	max_lsize_width = MAX(max_lsize_width, m);
    }
}


static void
getWinSize (VOID_ARG)
{
    /* On SunOS, tgetnum("li") returns the true window size, but under */
    /* ULTRIX, it returns the value from the termcap entry, which is usually */
    /* wrong in a window environment.  Get the true values if possible. */
    Tscrlen = tgetnum ("li") - 1;	/* Take defaults */
    scrwid = tgetnum ("co") - 1;	/* Don't scroll in last line */
#if defined(TIOCGWINSZ)
    {
	struct winsize win;

	if (ioctl (0, TIOCGWINSZ, &win) == 0)
	{
	    if (win.ws_col)
		scrwid = win.ws_col - 1;
	    if (win.ws_row)
		Tscrlen = win.ws_row - 1;
	}
    }
#endif
    if (Tscrlen <= 0 || scrwid <= 0)
    {
	FPRINTF (stderr, "Must know the screen size, assuming 24x80\n");
	Tscrlen = 24 - 1;
	scrwid = 80 - 1;
    }
    if (scrwid > MAXWIDTH)
	scrwid = MAXWIDTH;

#if defined(ultrix)
    scrwid--;			/* ULTRIX wraps in last column, safety factor 1 */
    Tscrlen--;			/* don't use final line, in case it is terminal status */
    /* line (e.g. line 25 on VT100 display) */
    /* ULTRIX scrolls when cursor enters last line */
#endif
}


static void
grow_file_array(VOID_ARG)
{
    if (file == (struct lbuf*) NULL)
	file = (struct lbuf*) malloc(FILE_SLOTS_TO_ADD * sizeof(struct lbuf));
    else
	file = (struct lbuf*) realloc((VOIDP)file,
				    (maxtotfiles + FILE_SLOTS_TO_ADD) *
				    sizeof(struct lbuf));
    if (file == (struct lbuf*) NULL)
	overflow();
    maxtotfiles += FILE_SLOTS_TO_ADD;
}


#if STDC
static int
gstat (const char *gname)
#else /* NOT STDC */
static int
gstat (gname)			/* Stats the file with name */
const char *gname;
#endif /* STDC */
{
    struct stat statbuf;

    file[totfiles].lflags = 0;
    file[totfiles].lnum = 0;
    file[totfiles].ltype = '-';

    if (show_link_target)
    {
	if (lstat (gname, &statbuf) < 0)
	    return (-1);
    }
    else
    {
	if (stat (gname, &statbuf) < 0)
	    return (-1);
    }

#if defined(HAVE_GRP_H)
    if (have_gid && (Gid != (uid_t) statbuf.st_gid))
	return (-1);		/* not requested group */
#endif

#if defined(HAVE_PWD_H)
    if (have_uid && (Uid != (uid_t) statbuf.st_uid))
	return (-1);		/* not requested user */
#endif

    file[totfiles].lnum = statbuf.st_ino;
    file[totfiles].lsize = statbuf.st_size;
    switch (statbuf.st_mode & S_IFMT)
    {
#if defined(S_IFDIR)
    case S_IFDIR:
	file[totfiles].ltype = 'd';
	break;
#endif

#if defined(S_IFBLK)
    case S_IFBLK:
	file[totfiles].ltype = 'b';
	file[totfiles].lsize = statbuf.st_rdev;
	break;
#endif

#if defined(S_IFCHR)
    case S_IFCHR:
	file[totfiles].ltype = 'c';
	file[totfiles].lsize = statbuf.st_rdev;
	break;
#endif

#if defined(S_IFDOOR)
    case S_IFDOOR:
	file[totfiles].ltype = 'D';
	break;
#endif

#if defined(S_IFLNK)
    case S_IFLNK:
	file[totfiles].ltype = 'l';
	break;
#endif

#if defined(S_IFIFO)
    case S_IFIFO:
	file[totfiles].ltype = 'p';
	break;
#endif

#if defined(S_IFSOCK)
    case S_IFSOCK:
	file[totfiles].ltype = 's';
	break;
#endif

    }
    file[totfiles].deleted = NOTDELETED;
    file[totfiles].lflags = statbuf.st_mode & ~S_IFMT;
    file[totfiles].luid = statbuf.st_uid;
    file[totfiles].lgid = statbuf.st_gid;
    file[totfiles].lnl = statbuf.st_nlink;
    file[totfiles].latime = statbuf.st_atime;
    file[totfiles].lctime = statbuf.st_ctime;
    file[totfiles].lmtime = statbuf.st_mtime;
    totblocks += nblock (statbuf.st_size);

    return (0);
}


#if STDC
static char *
makename (const char *dir, const char *filen, int len)
#else /* NOT STDC */
static char *
makename (dir, filen, len)
const char *dir;
const char *filen;		/* NB: might NOT be NUL terminated */
int len;
#endif /* STDC */
{
    static char dfile[MAXFILEPATHLEN];
    register char *dp;
    register const char *fp;
    register int i;

    dp = dfile;
    fp = dir;
    while (*fp)
	*dp++ = *fp++;
    *dp++ = '/';
    fp = filen;
    for (i = 0; i < len; i++)
    {
	*dp++ = *fp++;
	if (!*fp)
	    break;
    }
    *dp = '\0';			/* ensure NUL termination */
    return (dfile);
}


#if STDC
static void
my_exit (int status)
#else /* NOT STDC */
static void
my_exit (status)
int status;
#endif /* STDC */
{
    if (file != (struct lbuf*)NULL)
	free((VOIDP)file);
    exit (status);
}


#if STDC
static long
nblock (long size)
#else /* NOT STDC */
static long
nblock (size)
long size;
#endif /* STDC */
{
    return ((size + 511) >> 9);
}


#if STDC
static void
pentry (int whichone)
#else /* NOT STDC */
static void
pentry (whichone)
int whichone;
#endif /* STDC */
{
    register int i;
    int cur_x;
    register char *cp;
    register char *org_cp;

#if defined(USE_LS_DATE)
    static TIME_T six_months_ago = 0L;
#endif

    if (file[whichone].lnum == (unsigned long) (-1))
	return;
    cursor_x = 0;
    if (iflg)
    {
	PRINTF ("%5lu ", (unsigned long) file[whichone].lnum);
	cursor_x += 5 + 1;
    }
    if (sflg)
    {
	PRINTF ("%4ld ", nblock (file[whichone].lsize));
	cursor_x += 4 + 1;
    }
    if (lflg)
    {
	TIME_T the_time;

	/* Output line looks like
	   -rwxr-xr-x 2 beebe          118212 1997.08.28 09:03:41  gcc
	   or, with inode sorting,
	      435401 -rwxr-xr-x 2 beebe          118212 1997.08.28 09:03:41  gcc
	*/

	/* Optionally print inode field */
	if (sortyp == INODE)	/* show inodes only when sorting by them */
	{
	    PRINTF ("%11lu ", (unsigned long) file[whichone].lnum);
	    cursor_x += 11 + 1;
	}

	/* Print type and protection fields */
	putchar (file[whichone].ltype);
	cursor_x++;

	pmode (file[whichone].lflags);
	cursor_x += 9;

	/* Print link count */
	PRINTF ("%2lu ", (unsigned long) file[whichone].lnl);
	cursor_x += 2 + 1;

	/* Print either group or user, lexicographic or numeric */
	switch (sortyp)
	{
	case GIDNAME:
	    PRINTF ("%-*.*s ", max_gname_width, max_gname_width, file[whichone].groupname);
	    cursor_x += max_gname_width + 1;
	    break;

	case GIDNUMBER:
	    PRINTF ("%*ld ", max_gid_width, (long)file[whichone].lgid);
	    cursor_x += max_gid_width + 1;
	    break;

	default:			/* all other sort types get username */
	case UIDNAME:
	    PRINTF ("%-*.*s ", max_uname_width, max_uname_width, file[whichone].username);
	    cursor_x += max_uname_width + 1;
	    break;

	case UIDNUMBER:
	    PRINTF ("%*ld ", max_uid_width, (long)file[whichone].luid);
	    cursor_x += max_uid_width + 1;
	    break;
	}

	/* Print size field or for block and character devices, major,minor fields */
	if (file[whichone].ltype == 'b' || file[whichone].ltype == 'c')
	{
	    PRINTF ("%3ld,%3ld", (long) major ((int) file[whichone].lsize),
		    (long) minor ((int) file[whichone].lsize));
	    cursor_x += 3 + 1 + 3;
	}
	else
	{
	    PRINTF ("%*ld", max_lsize_width, file[whichone].lsize);
	    cursor_x += max_lsize_width;
	}

	/* Print one of read/create/write time */
	if (sortyp == READ)
	    the_time = file[whichone].latime;
	else if (sortyp == CREATE)
	    the_time = file[whichone].lctime;
	else			/* all other sort types: use last modification date */
	    the_time = file[whichone].lmtime;

#if defined(USE_LS_DATE)	/* use "Jul  4 12:32" or "Oct  9  1997" */
	if (six_months_ago == 0)	/* first time only */
	{
	    TIME_T ttime;

	    time (&ttime);
	    six_months_ago = ttime - 6L * 30L * 24L * 60L * 60L;
	}

	cp = ctime (&the_time);	/* returns "Sat Jul  4 13:16:03 MDT 1998\n" */
	if (the_time < six_months_ago)
	{
	    PRINTF (" %-7.7s %-4.4s ", cp + 4, cp + 20);
	    cursor_x += 1 + 7 + 1 + 4 + 1;
	}
	else
	{
	    PRINTF (" %-12.12s ", cp + 4);
	    cursor_x += 1 + 12 + 1;
	}
#else /* use ISO style date: "1998.07.04 12:32:59" */
	{
	    struct tm *t;

	    t = localtime (&the_time);
	    PRINTF (" %4d.%02d.%02d %02d:%02d:%02d ",
		    t->tm_year + 1900,
		    t->tm_mon + 1,
		    t->tm_mday,
		    t->tm_hour,
		    t->tm_min,
		    t->tm_sec);
	}
	cursor_x += 1 + 4 + 1 + 2 + 1 + 2 + 1 + 2 + 1 + 2 + 1 + 2 + 1;
#endif
    }

    /* cursor_x now points to the deletion flag position, just before the filename */

    /* Print deletion flag */
    PRINTF ("%c", file[whichone].deleted ? 'D' : ' ');

    PRINTF ("%c", ' ');	    /* dired 4.05: separate deletion flag from filename */

    /* Print optionally colorized filename */
    if (file[whichone].lflags & ISARG)
	org_cp = cp = file[whichone].ln.namep;
    else
	org_cp = cp = file[whichone].ln.lname;
    {					/* New for dired 4.00: colorize filenames */
	char buffer[2 * MAX_DNAME];	/* big enough for colorizing wrappers on filename */
	register char *bp;

	bp = &buffer[0];
	for (i = 0; (*cp && ((CURSOR + 1 + i) < scrwid)); i++)
	    *bp++ = ((*cp < ' ') || (*cp == '\177')) ? '?' : *cp++;
	*bp = '\0';
	cur_x = CURSOR + 1 + i + 1;

	(void) fputs (((batch_mode || monochrome) ?
		       buffer : colorize (buffer, &file[whichone])),
		      stdout);
	if (show_link_target && (file[whichone].ltype == 'l')) /* New for dired 4.05: show link target */
	{
	    char path[MAX_DNAME];

	    path[0] = '\0';
	    if (fullpath[0] != '\0')
	    {
		(void) strncat(path, fullpath, sizeof(path) - strlen(path));
		(void) strncat(path, "/", sizeof(path) - strlen(path));
	    }
	    (void) strncat(path, org_cp, sizeof(path) - strlen(path));
	    i = readlink(path, buffer, sizeof(buffer));
	    if (i > 0)
	    {
		buffer[i] = '\0';	/* readlink() doesn't NUL-terminate string */
		putstr(" -> ", &cur_x);
		putstr(buffer, &cur_x);
	    }
	}
    }
}


#if STDC
static void
putstr(char *s, int *pcur_x)
#else /* NOT STDC */
static void
putstr(s, pcur_x)
char *s;
int *pcur_x;
#endif /* STDC */
{
    for (; *s && (*pcur_x <= scrwid); ++s, (*pcur_x)++)
	fputc((int)(*s), stdout);
}


#if STDC
static int
getname (Boolean want_group, long gid_or_uid, char buf[], size_t maxbuf)
#else /* NOT STDC */
static int
getname (want_group, gid_or_uid, buf, maxbuf)
Boolean want_group;
long gid_or_uid;
char buf[];
size_t maxbuf;
#endif /* STDC */
{
    static char last_buf[MAX(MAX_USERNAME,MAX_GROUPNAME) + 1];
    static long last_gid_or_uid = UNSET_GID_OR_UID;
    static int last_want_group = -1;

#if defined(HAVE_GETGRGID) && defined(HAVE_GETPWUID) && defined(HAVE_PWD_H) && defined(HAVE_GRP_H)
    /* modern style password/group file search */

    if ((gid_or_uid == last_gid_or_uid) && (last_want_group == (int)want_group))
    {				/* important economization: avoid expensive */
	/* file lookup when gid/uid has not changed */
	(void) strncpy (buf, last_buf, MIN(maxbuf,sizeof (last_buf)));
	buf[maxbuf - 1] = '\0';
	return (0);
    }

    last_gid_or_uid = UNSET_GID_OR_UID;

    buf[0] = '\0';		/* clear buffer */
    if (want_group)		/* then search groupname list */
    {
	struct group *g;

	if ((g = getgrgid ((gid_t) gid_or_uid), g) != (struct group *) NULL)
	    (void) strncpy (buf, g->gr_name, MIN(maxbuf,strlen (g->gr_name)+1));
    }
    else
	/* search username list */
    {
	struct passwd *u;

	if ((u = getpwuid ((uid_t) gid_or_uid), u) != (struct passwd *) NULL)
	    (void) strncpy (buf, u->pw_name, MIN(maxbuf,strlen (u->pw_name)+1));
    }

    buf[maxbuf - 1] = '\0';	/* ensure string termination */
    last_gid_or_uid = gid_or_uid;	/* remember for next time around */
    last_want_group = (int)want_group;
    (void) strncpy (last_buf, buf, MIN(maxbuf, sizeof (last_buf)));
    last_buf[sizeof (last_buf) - 1] = '\0';

    return ((buf[0] == '\0') ? -1 : 0);

#else /* ancient style: read /etc/passwd and /etc/group directly */
    long n_gid_or_uid;

    static int first = 1;
    static FILE *pwdf = (FILE *) NULL;	/* Password/group file and directory files */

    if (first)
    {
	if (lflg)
	    pwdf = fopen ((want_group ? "/etc/group" : "/etc/passwd"), "r");
	first = 0;
    }

    if ((gid_or_uid == last_gid_or_uid) && (last_want_group == (int)want_group))
    {				/* important economization: avoid expensive */
	/* file lookup when gid/uid has not changed */
	(void) strncpy (buf, last_buf, MIN(maxbuf,sizeof (last_buf)));
	buf[maxbuf - 1] = '\0';
	return (0);
    }

    if (pwdf == NULL)		/* no password file available */
	return (-1);

    rewind (pwdf);
    buf[0] = '\0';
    last_gid_or_uid = UNSET_GID_OR_UID;
    do
    {
	int c;			/* current char from password file */
	size_t i;		/* index into buf[] */
	int j;			/* field number in password file entry:
				   "username:passwd:uid:gid:age:comment:gecos:homedir:shell" */

	i = 0;
	j = 0;
	n_gid_or_uid = 0L;
	while ((c = fgetc (pwdf)) != '\n')
	{
	    if (c == EOF)
		return (-1);	/* failure: no name found */
	    if (c == ':')	/* then end of current field */
	    {
		j++;
		c = '0';
	    }
	    if ((j == 0) && (i < maxbuf))	/* then collect username field */
		buf[i++] = c;
	    if (j == 2)
		n_gid_or_uid = n_gid_or_uid * 10 + (c - '0');
	}
	if (i > (maxbuf - 1))
	    i = maxbuf - 1;
	buf[i] = '\0';		/* terminate username */
    }
    while (n_gid_or_uid != gid_or_uid);
    last_gid_or_uid = gid_or_uid;
    last_want_group = (int)want_group;
    (void) strncpy (last_buf, buf, MIN(maxbuf, sizeof (last_buf)));
    last_buf[sizeof (last_buf) - 1] = '\0';

    return (0);			/* success */
#endif /* defined(HAVE_GETGRGID) && defined(HAVE_GETPWUID) && defined(HAVE_PWD_H) && defined(HAVE_GRP_H) */
}


#if STDC
static void
pmode (int aflag)
#else /* NOT STDC */
static void
pmode (aflag)
int aflag;
#endif /* STDC */
{
    register int **mp;

    flags = aflag;
    for (mp = &m[0]; mp < &m[sizeof (m) / sizeof (m[0])];)
	Select (*mp++);
}


#if STDC
static void
Select (register int *pairp)
#else /* NOT STDC */
static void
Select (pairp)
register int *pairp;
#endif /* STDC */
{
    register int n;

    n = *pairp++;
    while (--n >= 0 && (flags & *pairp++) == 0)
	pairp++;
    putchar (*pairp);
}


#define CMP_COMMON() \
	register const struct lbuf *p1; \
	register const struct lbuf *p2; \
	register int result; \
	p1 = (const struct lbuf *) pp1; \
	p2 = (const struct lbuf *) pp2; \
	if (p1->lflags & ISARG && p1->ltype == 'd') \
	{ \
	    if (!(p2->lflags & ISARG && p2->ltype == 'd')) \
		return (1); \
	} \
	else \
	{ \
	    if (p2->lflags & ISARG && p2->ltype == 'd') \
		return (-1); \
	}


#if STDC
static int
cmp_create (const void *pp1, const void *pp2)
#else /* NOT STDC */
static int
cmp_create (pp1, pp2)
const void *pp1;
const void *pp2;
#endif /* STDC */
{
    CMP_COMMON ();

    if (p2->lctime == p1->lctime)
	result = cmp_secondary (pp1, pp2);	/* equal primary key: compare secondary */
    else if (p2->lctime > p1->lctime)	/* default: descending */
	result = 1;
    else
	result = -1;
    return (rflg * result);
}


#if STDC
static int
cmp_gid (const void *pp1, const void *pp2)
#else /* NOT STDC */
static int
cmp_gid (pp1, pp2)
const void *pp1;
const void *pp2;
#endif /* STDC */
{					/* compare group names, or else group numbers */
    CMP_COMMON ();

    if (p2->lgid == p1->lgid)
	result = cmp_secondary (pp1, pp2); /* equal primary key: compare secondary */
    else if (p2->groupname[0] && p1->groupname[0]) /* compare group names */
	result = strcmp (p1->groupname, p2->groupname);
    else if (p2->lgid < p1->lgid)	/* compare group numbers */
	result = 1;
    else
	result = -1;
    return (rflg * result);
}


#if STDC
static int
cmp_gid_number (const void *pp1, const void *pp2)
#else /* NOT STDC */
static int
cmp_gid_number (pp1, pp2)
const void *pp1;
const void *pp2;
#endif /* STDC */
{					/* compare group numbers */
    CMP_COMMON ();

    if (p2->lgid == p1->lgid)
	result = cmp_secondary (pp1, pp2); /* equal primary key: compare secondary */
    else if (p2->lgid < p1->lgid)	/* compare group numbers */
	result = 1;
    else
	result = -1;
    return (rflg * result);
}


#if STDC
static int
cmp_inode (const void *pp1, const void *pp2)
#else /* NOT STDC */
static int
cmp_inode (pp1, pp2)
const void *pp1;
const void *pp2;
#endif /* STDC */
{
    CMP_COMMON ();

    if (p2->lnum == p1->lnum)
	result = cmp_secondary (pp1, pp2);	/* equal primary key: compare secondary */
    else if (p2->lnum < p1->lnum)
	result = 1;
    else
	result = -1;
    return (rflg * result);
}


#if STDC
static int
cmp_links (const void *pp1, const void *pp2)
#else /* NOT STDC */
static int
cmp_links (pp1, pp2)
const void *pp1;
const void *pp2;
#endif /* STDC */
{
    CMP_COMMON ();

    if (p2->lnl == p1->lnl)
	result = cmp_secondary (pp1, pp2);	/* equal primary key: compare secondary */
    else if (p2->lnl > p1->lnl)	/* default: descending */
	result = 1;
    else
	result = -1;
    return (rflg * result);
}


#if STDC
static int
cmp_name (const void *pp1, const void *pp2)
#else /* NOT STDC */
static int
cmp_name (pp1, pp2)
const void *pp1;
const void *pp2;
#endif /* STDC */
{					/* compare filenames lexicographically */
    CMP_COMMON ();

    return (rflg * strcmp (p1->lflags & ISARG ? p1->ln.namep : p1->ln.lname,
			 p2->lflags & ISARG ? p2->ln.namep : p2->ln.lname));
}


#if STDC
static int
cmp_number (const void *pp1, const void *pp2)
#else /* NOT STDC */
static int
cmp_number (pp1, pp2)
const void *pp1;
const void *pp2;
#endif /* STDC */
{					/* compare filenames as numbers */
    register long n1;
    register long n2;
    register const char *q1;
    register const char *q2;
    char *endptr1;
    char *endptr2;
    Boolean numeric_sort;

    CMP_COMMON ();

    q1 = p1->lflags & ISARG ? p1->ln.namep : p1->ln.lname;
    q2 = p2->lflags & ISARG ? p2->ln.namep : p2->ln.lname;

#if defined(HAVE_STRTOL)
    n1 = strtol (q1, &endptr1, 0);
    n2 = strtol (q2, &endptr2, 0);
    numeric_sort = ((endptr1 != (const char*) NULL) && (*endptr1 == '\0') &&
		    (endptr2 != (const char*) NULL) && (*endptr2 == '\0'))
		    ? True: False;
#else
    /* When strtol() is unavailable, fall back to the unchecked atoi()
       function, sigh... */
    n1 = (long) atoi (q1);
    n2 = (long) atoi (q2);
    numeric_sort = True;
#endif

    if (numeric_sort)			/* should have found valid numbers */
    {
	if (n1 == n2)
	    result = cmp_secondary (pp1, pp2);	/* equal primary key: compare secondary */
	else if (n1 < n2)
	    result = -1;
	else
	    result = 1;
	return (rflg * result);
    }
    else	/* fall back to lexicographic sort in current sort order */
	return (cmp_name (pp1,pp2));
}


#if STDC
static int
cmp_read (const void *pp1, const void *pp2)
#else /* NOT STDC */
static int
cmp_read (pp1, pp2)
const void *pp1;
const void *pp2;
#endif /* STDC */
{
    CMP_COMMON ();

    if (p2->latime == p1->latime)
	result = cmp_secondary (pp1, pp2);	/* equal primary key: compare secondary */
    else if (p2->latime > p1->latime)	/* default: descending */
	result = 1;
    else
	result = -1;
    return (rflg * result);
}


#if STDC
static int
cmp_secondary (const void *pp1, const void *pp2)
#else /* NOT STDC */
static int
cmp_secondary (pp1, pp2)
const void *pp1;
const void *pp2;
#endif /* STDC */
{
    /* NB: When rflg == -1, both cmp_name() and the caller will reverse
       the sense of the comparison: the net effect is that the secondary
       sort for equal primary keys will always be in ASCENDING order,
       which is documented, and intended. */
    return (cmp_name (pp1, pp2));
}


#if STDC
static int
cmp_size (const void *pp1, const void *pp2)
#else /* NOT STDC */
static int
cmp_size (pp1, pp2)
const void *pp1;
const void *pp2;
#endif /* STDC */
{
    CMP_COMMON ();

    if (p2->lsize == p1->lsize)
	result = cmp_secondary (pp1, pp2);	/* equal primary key: compare secondary */
    else if (p2->lsize > p1->lsize)	/* default: descending */
	result = 1;
    else
	result = -1;
    return (rflg * result);
}


#if STDC
static int
cmp_type (const void *pp1, const void *pp2)
#else /* NOT STDC */
static int
cmp_type (pp1, pp2)
const void *pp1;
const void *pp2;
#endif /* STDC */
{
    CMP_COMMON ();

    if (p2->ltype == p1->ltype)
	result = cmp_secondary (pp1, pp2);	/* equal primary key: compare secondary */
    else if (p2->ltype < p1->ltype)
	result = 1;
    else
	result = -1;
    return (rflg * result);
}


#if STDC
static int
cmp_uid (const void *pp1, const void *pp2)
#else /* NOT STDC */
static int
cmp_uid (pp1, pp2)
const void *pp1;
const void *pp2;
#endif /* STDC */
{					/* compare user names, or else numbers */
    CMP_COMMON ();

    if (p2->luid == p1->luid)
	result = cmp_secondary (pp1, pp2); /* equal primary key: compare secondary */
    else if (p2->username[0] && p1->username[0]) /* compare user names */
	result = strcmp (p1->username, p2->username);
    else if (p2->luid < p1->luid)	/* compare user numbers */
	result = 1;
    else
	result = -1;
    return (rflg * result);
}


#if STDC
static int
cmp_uid_number (const void *pp1, const void *pp2)
#else /* NOT STDC */
static int
cmp_uid_number (pp1, pp2)
const void *pp1;
const void *pp2;
#endif /* STDC */
{					/* compare user numbers */
    CMP_COMMON ();

    if (p2->luid == p1->luid)
	result = cmp_secondary (pp1, pp2);	/* equal primary key: compare secondary */
    else if (p2->luid < p1->luid)
	result = 1;
    else
	result = -1;
    return (rflg * result);
}


#if STDC
static int
cmp_write (const void *pp1, const void *pp2)
#else /* NOT STDC */
static int
cmp_write (pp1, pp2)
const void *pp1;
const void *pp2;
#endif /* STDC */
{
    CMP_COMMON ();

    if (p2->lmtime == p1->lmtime)
	result = cmp_secondary (pp1, pp2);	/* equal primary key: compare secondary */
    else if (p2->lmtime > p1->lmtime)	/* default: descending */
	result = 1;
    else
	result = -1;
    return (rflg * result);
}


#if 0				/* UNUSED */
static void
ceod (VOID_ARG)
{
}
#endif


#if STDC
static void
ceol (int x, int y)
#else /* NOT STDC */
static void
ceol (x, y)
int x;
int y;
#endif /* STDC */
{
    int i;

    if (my_CE != (char*) NULL)
	putpad (my_CE);
    else		/***crh for terms w/o ce, put blanks & restore cursor **/
    {
	for (i = x; i < scrwid; i++)
	    putchar (' ');
	curxy (x, y);
    }
}


static void
blank (VOID_ARG)
{
    putpad (my_CL);
}


static void
help_and_exit (VOID_ARG)
{
    FPRINTF(stderr, "Usage:\n");
    FPRINTF(stderr, "\n");
    FPRINTF(stderr, "    %s [ -? ] [ -b ] [ -d ] [ -ffirstfile ] [ -Ffirstfile ]\n", prog);
    FPRINTF(stderr, "          [ -g[gid-or-groupname] ] [ -h ] [ -m ]\n");
    FPRINTF(stderr, "          [ -[s|r][c|g|G|i|l|n|N|r|s|t|u|U|w] ] [ -S ]\n");
    FPRINTF(stderr, "          [ -u[uid-or-username] ] [ -v ] [ -V ] [ -w[f|h|number] ]\n");
    FPRINTF(stderr, "          [ directory-list|file-list ]\n");
    FPRINTF(stderr, "\n");
    FPRINTF(stderr, "    where the options are:\n");
    FPRINTF(stderr, "\n");
    FPRINTF(stderr, "    -?        display this help message and quit\n");
    FPRINTF(stderr, "    -b        batch mode\n");
    FPRINTF(stderr, "    -d        subprocess flag (internal use only)\n");
    FPRINTF(stderr, "    -ffile    start display with this file\n");
    FPRINTF(stderr, "    -Ffile    same as -ffile\n");
    FPRINTF(stderr, "    -gid      select files belonging to group id (a name or number)\n");
    FPRINTF(stderr, "    -h        display this help message and quit\n");
    FPRINTF(stderr, "    -m        monochrome mode\n");
    FPRINTF(stderr, "    -rx       reverse sort by key x (see -sx)\n");
    FPRINTF(stderr, "    -sx       sort by key x (one of c, g, G, i, l, n, N, r, s, t, u, U,\n");
    FPRINTF(stderr, "              or w), for creation date, group name, group number, link\n");
    FPRINTF(stderr, "              count, name (lexicographic), name (numeric), read date,\n");
    FPRINTF(stderr, "              size, type, user name, user number, or write date\n");
    FPRINTF(stderr, "              respectively\n");
    FPRINTF(stderr, "    -S        suppress display of symbolic-link targets\n");
    FPRINTF(stderr, "    -uid      select files belonging to user id (a name or number)\n");
    FPRINTF(stderr, "    -v        show program version on stderr and quit\n");
    FPRINTF(stderr, "    -V        same as -v\n");
    FPRINTF(stderr, "    -wf       use full-screen display (default)\n");
    FPRINTF(stderr, "    -wh       use half-screen display\n");
    FPRINTF(stderr, "    -wnumber  use number lines for the display\n");

    my_exit(EXIT_SUCCESS);
}


static void
home (VOID_ARG)
{
    curxy (0, 0);		/* crh putpad (my_HO); */
}


static void
insline (VOID_ARG)
{
    putpad (my_AL);
}


/* Yes, folks, we use direct cursor addressing to get to next line!
   Before you mumble "What sort of cretin would do this?" here's
   the reason. We don't use \n since that obviously won't work.
   We don't use \012 since virgin version 7 makes that into a crlf.
   We don't use raw mode since we type out help files efficently,
   and we don't want to switch modes all the time. So enjoy. -- SMC */

static void
downline (VOID_ARG)
{
    curxy (CURSOR, ++curline);
    curfile++;
}


#if 0				/* UNUSED */
static void
upline (VOID_ARG)
{
    putpad (my_UP);
    curline--;
    curfile--;
}
#endif


#if STDC
static void
telluser (const char *msg, const char *arg)
#else /* NOT STDC */
static void
telluser (msg, arg)
const char *msg;
const char *arg;
#endif /* STDC */
{
    curxy (0, Tscrlen);
    ceol (0, Tscrlen);
    PRINTF (msg, arg);
    curxy (CURSOR, curline);
    blurb++;
    (void) fflush (stdout);	/* force it out quickly */
}


#if STDC
static void
curxy (int col, int lin)
#else /* NOT STDC */
static void
curxy (col, lin)
int col;
int lin;
#endif /* STDC */
{
    if (!batch_mode)
    {
	putpad ((char*)tgoto (my_CM, col, lin));
	(void) fflush (stdout);
    }
}


#if STDC
static Boolean
nottext (const char fn[])
#else /* NOT STDC */
static Boolean
nottext (fn)
const char fn[];
#endif /* STDC */
{
    /*
     * Routine to try to determine whether the file is obviously not a
     * text file.
     *
     * Adapted from "more".   *** crh **
     *
     * dired 4.00: The old version had nasty machine-specific code
     * checking the first pair of bytes in the file for `magic'
     * numbers of PDP-11 executable file, and using access to internal
     * details of the FILE* structure.  This new version examines up
     * to the first MAXLOOKAHEAD characters of the file looking for
     * binary characters, and does so portably.
     */

    int c;
    FILE *fp;
    size_t k;

    if ((fp = fopen (fn, "r")) == NULL)
	return (False);		/* non-existent file */

    for (k = 0; (k++ < (size_t) MAXLOOKAHEAD) && (c = getc (fp), c) != EOF;)
    {
	if ((c < 0) || (c > 127))	/* should be isascii(c), but that */
	{			/* is not in Standard C */
	    (void) fclose (fp);
	    return (True);
	}

    }
    (void) fclose (fp);
    return (False);
}


#if STDC
static int
Getch (FILE * fd)
#else /* NOT STDC */
static int
Getch (fd)
FILE *fd;
#endif /* STDC */
{
    if (wasspecial)
    {
	wasspecial = False;
	return (specialcc);
    }
    else
	return (getc (fd));
}


#if STDC
static Boolean
type (const char *filestr, Boolean waitflg)
#else /* NOT STDC */
static Boolean
type (filestr, waitflg)		/* Modified to type help file & others. fjl 5/81 */
const char *filestr;		/* Kludgy now with split screen stuff! */
Boolean waitflg;
#endif /* STDC */
{
    FILE *fd;
    int cur_scrl;		/* current screen length */
    register int cc = 0;
    register int i;
    register int j;
    register int n;

    fd = stdin;

    if (nottext (filestr))
    {
	telluser ("?Not a text file:%s", filestr);
	return (False);
    }
    fd = fopen (filestr, "r");
    if (fd == (FILE*) NULL)
    {
	telluser ("?Unable to open %sfile",
		  (STRING_EQUAL (filestr, helpfile) ? "help " : ""));
	return (False);
    }

    SIGNAL (SIGINT, catchint);
    sigint = False;
    intype = True;

    if (!splitflg)
    {
	char helpbuf[BUFSIZ];

	blank ();
	(void) fflush (stdout);
	while (((i = fread (helpbuf, 1, sizeof(helpbuf), fd), i) > 0) && !sigint)
	    (void) fwrite (helpbuf, 1, i, stdout);
	(void) fwrite ("\n", 1, 1, stdout); /* extra newline to avoid later */
	(void) fclose (fd); /* overwriting last line by "CR to return..." message */
	(void) fflush (stdout);
    }
    else
    {
	cur_scrl = totfiles - topfile + 1;	/* topfile starts at 0 */
	Worktop = MIN(cur_scrl,scrlen) + 1;
	if (!splitflg)
	{
	    blank ();
	    Worktop = 0;
	}
	do
	{
	    curxy (0, Worktop); /**** crh  replace call to ceod()  ****/
	    if (my_CD != (char*) NULL)
		putpad (my_CD);
	    else if (!splitflg)
		blank ();
	    wasspecial = False;
	    for (i = Worktop; (i < Tscrlen) && !sigint && (cc != EOF); i++)
	    {
		n = 0;
		while ((cc = Getch (fd)) != EOF)
		{
		    if (cc == '\t')
			n = (n / 8) * 8 + 8;
		  chklen:
		    if (n >= scrwid /*+1 */ )
		    {		/* Use only 79 cols crh** */
			if (cc != '\n')
			{
			    if (!wasspecial)
				ungetc (cc, fd);
			    else
				specialcc = cc;
			    cc = '\n';
			}
		    }
		    if (cc >= 127)
		    {
			putchar ('^');
			cc = '?';
			n++;
			wasspecial = True;
			goto chklen;
		    }
		    else if ((cc >= 0) && (cc <= 31) && (cc != '\t') && (cc != '\n'))
		    {
			/* form the graphic of the control char */
			putchar ('^');
			cc = cc + 64;
			n++;
			wasspecial = True;
			goto chklen;
		    }
		    if (cc != '\n')
		    {
			putchar (cc);
			wasspecial = False;
		    }
		    else
		    {
			if ((my_CD == (char *) NULL) && splitflg)
			    ceol (n, i);
			putchar ('\n');
			break;
		    }
		    if (cc != '\t')
			n++;
		}
	    }
	    if ((my_CD == (char*) NULL) && splitflg)
	    {
		for (j = i; j < Tscrlen; j++)
		{
		    curxy (0, j);
		    ceol (0, j);
		}
	    }
	    (void) fflush (stdout);
	}
	while (!sigint && (cc != EOF) && waitchk (waitflg));

	if (feof (fd))
	    PRINTF ("===== End-of-File =====\n");
	(void) fflush (stdout);
	(void) fclose (fd);
	return (False);		/* needs no re-display */
    }

    if (!splitflg)
    {				/* redundant now... */
	if (!sigint)
	{
	    curxy (0, Tscrlen);
	    PRINTF ("CR to return...");
	    (void) fflush (stdout);
	    getchar ();
	}
	blank ();
	return (True);
    }
    return (False);
}


#if STDC
static int
waitchk (Boolean waitflg)
#else /* NOT STDC */
static int
waitchk (waitflg)
Boolean waitflg;
#endif /* STDC */
{
    if (!waitflg)
	return (1);
    (void) fflush (stdout);
    curxy (0, Tscrlen);
    ceol (0, Tscrlen);
    PRINTF ("---Continue---");
    (void) fflush (stdout);
    curxy (0, Tscrlen);
    if (getchar () == 'q')
	sigint = True;		/* simulate interrupt */
    ceol (0, Tscrlen);
    return (sigint ? 0 : 1);
}


#if defined(USE_TERMIOS)
static void
setdpy (VOID_ARG)
{
    struct termios tty;		/* XPG3, POSIX.1, FIPS 151-1 interface */

    if (tcgetattr (fileno (stdin), &tty) != -1)
    {
	tty_save = tty;
	tty.c_iflag &= ~(INLCR | ICRNL | ISTRIP | IXON | BRKINT);

#if defined(IUCLC)
	tty.c_iflag &= ~IUCLC;	/* absent from POSIX */
#endif /* defined(IUCLC) */

	tty.c_lflag &= ~(ECHO | ICANON);
	tty.c_cc[VMIN] = 5;	/* MIN */
	tty.c_cc[VTIME] = 2;	/* TIME */
	(void) tcsetattr (fileno (stdin), TCSANOW, &tty);
    }
}


static void
unsetdpy (VOID_ARG)
{
    (void) tcsetattr (fileno (stdin), TCSANOW, &tty_save);
}
#endif /* defined(USE_TERMIOS) */


#if defined(USE_TERMIO)
static void
setdpy (VOID_ARG)
{
    struct termio tty;		/* System V interface */

    if (tcgetattr (fileno (stdin), &tty) != -1)
    {
	tty_save = tty;
	tty.c_iflag &= ~(INLCR | ICRNL | ISTRIP | IXON | BRKINT);

#if defined(IUCLC)
	tty.c_iflag &= ~IUCLC;	/* absent from POSIX */
#endif /* defined(IUCLC) */

	tty.c_lflag &= ~(ECHO | ICANON);
	tty.c_cc[VMIN] = 5;	/* MIN */
	tty.c_cc[VTIME] = 2;	/* TIME */
	(void) tcsetattr (fileno (stdin), TCSANOW, &tty);
    }
}


static void
unsetdpy (VOID_ARG)
{
    (void) tcsetattr (fileno (stdin), TCSANOW, &tty_save);
}
#endif /* defined(USE_TERMIO) */


#if defined(USE_SGTTY)
static void
setdpy (VOID_ARG)
{
    ioctl (0, TIOCGETP, &ioctlb);
    ioctlb.sg_flags |= CBREAK;
    ioctlb.sg_flags &= ~ECHO;
    ioctl (0, TIOCSETP, &ioctlb);
}


static void
unsetdpy (VOID_ARG)
{
    ioctlb.sg_flags &= ~CBREAK;
    ioctlb.sg_flags |= ECHO;
    ioctl (0, TIOCSETP, &ioctlb);
}
#endif /* defined(USE_SGTTY) */

#if STDC
static void
set_sorttyp(int direction, int type)
#else /* NOT STDC */
static void
set_sorttyp(direction, type)
int direction;		/* 's' or 'r' (only 's' is checked for) */
int type;		/* one of the characters in sortstr[] */
#endif /* STDC */
{			/* set sortyp, defaulting to NAME if type is invalid */
    const char *p;

    p = (const char*) strchr (sortstr, type);
    sortyp = (p == (const char*)NULL) ? NAME : (int) (p - sortstr);
    cmp_fcn = cmp_table[sortyp];
    rflg = (direction == 's') ? 1 : -1;
}


static void
Getcap (VOID_ARG)
{
    char *ap;
    char *term;
    char tbuf[1024];
    static char tcapbuf[256];	/* must be static since the my_XX caps point into it */

#if 0
    char *xPC;
#endif

    term = getenv ("TERM");
    if (term == (char*) NULL)
    {
	FPRINTF (stderr, "No TERM in environment\n");
	my_exit (EXIT_FAILURE);
    }

    switch ((int) (tgetent (tbuf, term)))
    {
    case -1:
	FPRINTF (stderr, "Cannot open termcap file\n");
	my_exit (2);
    case 0:
	FPRINTF (stderr, "%s: unknown terminal", term);
	my_exit (3);
    }

    ap = tcapbuf;

    Tscrlen = tgetnum ("li") - 1;
    scrwid = tgetnum ("co") - 1;	/* lose 1 so won't scroll in last line */

#if defined(TIOCGWINSZ)
    /* On SunOS, tgetnum("li") returns the true window size, but under
       ULTRIX, it returns the value from the termcap entry, which is usually
       wrong in a window environment.  Get the true values if possible. */
    {
	struct winsize win;

	if (ioctl (0, TIOCGWINSZ, &win) == 0)
	{
	    if (win.ws_col)
		scrwid = win.ws_col - 1;
	    if (win.ws_row)
		Tscrlen = win.ws_row - 1;
	}
    }
#endif

    if (scrwid > MAXWIDTH)
	scrwid = MAXWIDTH;

#if defined(ultrix)
    scrwid--;			/* ULTRIX wraps in last column */
    Tscrlen--;			/* ULTRIX scrolls when cursor */
    /* enters last line in */
    /* full-window mode, so */
    /* shorten the height by 1 line */
#endif


    my_UP = tgetstr ("up", &ap);
    my_CD = tgetstr ("cd", &ap);
    my_CE = tgetstr ("ce", &ap);

#if 0
    my_HO = tgetstr ("ho", &ap);
#endif

    my_CL = tgetstr ("cl", &ap);
    my_CM = tgetstr ("cm", &ap);
    my_AL = tgetstr ("al", &ap);	/* insert line, optional */

#if 0
    xPC = tgetstr ("pc", &ap);
    if (xPC)
	my_PC = *xPC;
#endif

    if ((my_CM == (char*) NULL) || (my_CL == (char*) NULL) || (my_UP == (char*) NULL))
    {				/*** crh changed error test ***/
	FPRINTF (stderr, "Tty must have cursor addr, clear, and 4 cursor motions.\n");
	my_exit (EXIT_FAILURE);
    }
    if (Tscrlen <= 0 || scrwid <= 0)
    {
	FPRINTF (stderr, "Must know the screen size\n");
	my_exit (EXIT_FAILURE);
    }
}


#if STDC
static int
outch (OUTCH_ARGTYPE c)
#else
static int
outch (c)
OUTCH_ARGTYPE c;
#endif /* STDC */
{
    return (putchar (c));
}


#if STDC
static void
putpad (char *str)
#else /* NOT STDC */
static void
putpad (str)
char *str;
#endif /* STDC */
{
    if (!batch_mode && (str != (char *) NULL))
	tputs (str, 0, outch);
}


#if STDC
static handler_t
catchint (int sig)
#else /* NOT STDC */
static handler_t
catchint (sig)
int sig;
#endif /* STDC */
{
    sig = sig;			/* avoid compiler unused-argument warnings */
    SIGNAL (SIGINT, SIG_IGN);	/* reset it */
    sigint = True;
#if !defined(HAVE_VOID_HANDLER)
    return (0);
#endif
}


#if STDC
static char *
bldnam (char *str, int narg, int filidx, char *argv[])
#else /* NOT STDC */
static char *
bldnam (str, narg, filidx, argv)
char *str;
int narg;
int filidx;
char *argv[];
#endif /* STDC */
{

    if (narg == 1)
	strcat (str, file[filidx].ln.lname);
    else if (narg == 2)
    {
	if (arglist)
	    strcat (str, file[filidx].ln.namep);
	else
	{
	    strcat (str, argv[1]);
	    strcat (str, "/");
	    strcat (str, file[filidx].ln.lname);
	}
    }
    else
	strcat (str, file[filidx].ln.namep);
    return (str);
}


#if STDC
static char *
catargs (char *str, char *argv[])
#else /* NOT STDC */
static char *
catargs (str, argv)
char *str;
char *argv[];
#endif /* STDC */
{
    char *cp;

    for (++argv; *argv; argv++)
    {
	if (**argv == '-')
	{
	    cp = &(**argv) + 1;
	    if (*cp != 'f' && *cp != 'F' && *cp != 'd')
	    {
		strcat (str, " ");
		strcat (str, *argv);
	    }
	}
    }
    return (str);
}


static void
overflow (VOID_ARG)
{
    PRINTF ("\n?Too many files\007\n");
    (void) fflush (stdout);
    unsetdpy ();
    sleep (1);			/* So user can see it - don't worry if less */
    my_exit (EXIT_FAILURE);
}


#if STDC
static char *
skipto (const char *string, const char *charset)
#else /* NOT STDC */
static char *
skipto (string, charset)
const char *string;
const char *charset;
#endif /* STDC */
{
    register const char *setp;
    register const char *strp;
    register Boolean found;

    found = False;		/* not found yet */
    strp = string;		/* start at first char */

    while (*strp && !found)
    {				/* until null or found */
	/* find first char in charset matching *strp */
	for (setp = charset; (*setp) && (*setp != *strp); setp++);
	if (*setp)
	    found = True;	/* matches a char */
	else
	    strp++;		/* else keep looking */
    }

    return ((char *) strp);
}
